"""Abstraction of the Daemon's logging functionality.
"""
import os 
import yaml
import logging
import inspect
from pathlib import Path
from logging.handlers import RotatingFileHandler


class Logger():

    DEBGUG_LEVEL = 10
    INFO_LEVEL = 20
    WARNING_LEVEL = 30
    ERROR_LEVEL = 40
    CRITICAL_LEVEL = 50

    def __init__(
            self,
            log_dirpath: str,
            log_max_size_bytes: int,
            log_file_count: int,
            log_message_source_path: str=None,
        ):
        """Initialize logger, acting as a proxy for pre-defined messages and logs on different logging levels.

        Notes
        -----
        Creates a 5-level log, consisting of info, debug, warning, error, and critical.

        Parameters
        ----------
        log_dirpath : str
            Path to where the logs will be produced.
        log_max_size_bytes : int
            Maximal size of an individual file in bytes.
        log_file_count : int
            Number of log files per log level (5 levels in total).
        log_message_source_path : str, optional
            Path to the definitions of pre-defined log messages. Default is None (log in src/valar_daemon dir).
        """
        # Load in pre-defined messages
        if log_message_source_path is None:
            log_message_source_path = str(Path(Path(__file__).parent, 'log_messages_source.yaml'))
        log_messages = Logger._fetch_log_messages(log_message_source_path)
        self.log_messages = Logger._strip_trailing_newline(log_messages)

        # Make logger master directory
        Logger.try_to_make_directory(log_dirpath)

        # Make debug subdirectory and initialize debug logger
        debug_log_dirpath = Path(log_dirpath, f'{self.DEBGUG_LEVEL}-debug')
        Logger.try_to_make_directory(debug_log_dirpath)
        self.debug_logger = self.create_logger(
            debug_log_dirpath,
            'debug', 
            log_max_size_bytes,
            log_file_count - 1 # If greater than 0, log backups are added
        )

        # Make info subdirectory and initialize info logger
        info_log_dirpath = Path(log_dirpath, f'{self.INFO_LEVEL}-info')
        Logger.try_to_make_directory(info_log_dirpath)
        self.info_logger = self.create_logger(
            info_log_dirpath,
            'info', 
            log_max_size_bytes,
            log_file_count - 1 # If greater than 0, log backups are added
        )

        # Make warning subdirectory and initialize info warning
        warning_log_dirpath = Path(log_dirpath, f'{self.WARNING_LEVEL}-warning')
        Logger.try_to_make_directory(warning_log_dirpath)
        self.warning_logger = self.create_logger(
            warning_log_dirpath,
            'warning', 
            log_max_size_bytes,
            log_file_count - 1 # If greater than 0, log backups are added
        )

        # Make error subdirectory and initialize info error
        error_log_dirpath = Path(log_dirpath, f'{self.ERROR_LEVEL}-error')
        Logger.try_to_make_directory(error_log_dirpath)
        self.error_logger = self.create_logger(
            error_log_dirpath,
            'error', 
            log_max_size_bytes,
            log_file_count - 1 # If greater than 0, log backups are added
        )

        # Make critical subdirectory and initialize info critical
        critical_log_dirpath = Path(log_dirpath, f'{self.CRITICAL_LEVEL}-critical')
        Logger.try_to_make_directory(critical_log_dirpath)
        self.critical_logger = self.create_logger(
            critical_log_dirpath,
            'critical', 
            log_max_size_bytes,
            log_file_count - 1 # If greater than 0, log backups are added
        )

    ### Utilities ######################################################################################################

    @staticmethod
    def try_to_make_directory(dirpath: str):
        """Try to make a directory and any parents if these are missing.

        Parameters
        ----------
        dirpath : str
            Path to the directory that needs to be created.

        Raises
        ------
        e
            An error, except `FileExistsError`.
        """
        try:
            os.makedirs(dirpath)
        except FileExistsError as e:
            pass
        except Exception as e:
            raise e 

    @staticmethod
    def create_logger(
        log_dirpath: str,
        logname: str,
        log_max_size_bytes: int,
        log_file_count: int
    ) -> logging.Logger:
        """Create logger.

        Parameters
        ----------
        log_dirpath : str
            Path to where the logs will be produced.
        logname : str
            Name of the created log.
        log_max_size_bytes : int
            Maximal size of an individual file in bytes.
        log_file_count : int
            Number of log files per log level (5 levels in total).

        Returns
        -------
        logging.Logger
            The created logger.
        """
        # Create log file handler
        handler = RotatingFileHandler(
            Path(log_dirpath, f'{logname}.log'), 
            maxBytes=log_max_size_bytes, 
            backupCount=log_file_count
        )
        # Set logging level - allow the logging of everything
        handler.setLevel(logging.DEBUG)
        # Set up the prefix of each message
        handler.setFormatter(logging.Formatter(
            '%(asctime)s - %(levelname)s - %(message)s'
        ))
        # Create logger
        logger = logging.getLogger(logname)
        # Set log level once more, this time for the logger
        logger.setLevel(logging.DEBUG)
        # Connect the log file handler to the logger 
        logger.addHandler(handler)
        return logger
        
    @staticmethod
    def _fetch_log_messages(log_message_source_path: str) -> dict:
        """Fetch the pre-defined log messages.

        Parameters
        ----------
        log_message_source_path : str
            Path to the definitions of pre-defined log messages.

        Returns
        -------
        dict
            Pre-defined log messages.
        """
        log_file = Path(log_message_source_path)
        with log_file.open("r") as f:
            return yaml.safe_load(f)
        
    @staticmethod
    def _strip_trailing_newline(log_messages: dict) -> dict:
        """Strip the pre-defined log messages of trailing `\n`, which are likely present, depending on user input.

        Parameters
        ----------
        log_messages : dict
            Pre-defined log messages.

        Returns
        -------
        dict
            Log messages without trailing `\n`
        """
        for _, value in log_messages.items():
            value['description'] = value['description'].rstrip("\n")
            value['message'] = value['message'].rstrip("\n")
        return log_messages

    @staticmethod
    def get_filename_and_lineno(call_stack_regression_levels: int=2) -> str:
        """Get the file name and line number string for prepending to log messages.

        Notes
        -----
        Calling through pre-defined messages requires 4-level regression and through the `info` etc. 2-level.
        Would be simpler to always call bottom-most caller (const. level), but it changes behavior when when debugging.

        Parameters
        ----------
        call_stack_regression_levels : int, optional
            Number of levels regresses on the call stack when lookin up the original file and line number, by default 2.

        Returns
        -------
        str
            Format accoring to: '<file>:<line>'
        """
        # caller_frame = inspect.currentframe().f_back.f_back
        caller_frame = inspect.stack()[call_stack_regression_levels]
        # filename = os.path.basename(caller_frame.f_code.co_filename)
        filename = os.path.basename(caller_frame.filename)
        # lineno = caller_frame.f_lineno
        lineno = caller_frame.lineno
        return f'{filename}:{lineno}'


    def _log(
            self, 
            level: int, 
            message: str, 
            **kwargs
        ):
        """Proxy for logging a message at the corresponding level.

        Parameters
        ----------
        level : int
            The logging level.
        message : str
            Logged message.
        """
        kwargs['call_stack_regression_levels'] = 5
        if level == self.DEBGUG_LEVEL:
            self.debug(message, **kwargs)
        elif level == self.INFO_LEVEL:
            self.info(message, **kwargs)
        elif level == self.WARNING_LEVEL:
            self.warning(message, **kwargs)
        elif level == self.ERROR_LEVEL:
            self.error(message, **kwargs)
        elif level == self.CRITICAL_LEVEL:
            self.critical(message, **kwargs)
        

    ### Generic messaging ##############################################################################################

    @staticmethod
    def prepend_filename_and_lineno_to_message(
        message: str,
        **kwargs
    ) -> str:
        """Prepend the name of the file and the file number where the logger was called.

        Parameters
        ----------
        message : str
            The logged message.

        Returns
        -------
        str
            Formatted message as: '<file and line> - <message>'.
        """
        return f'{Logger.get_filename_and_lineno(**kwargs)} - ' + message


    def debug(
            self, 
            message: str, 
            append_filename_and_lineno: bool=True, 
            **kwargs
        ):
        """Log a debug message.

        Parameters
        ----------
        message : str
            Logged message.
        append_filename_and_lineno : bool, optional
            Flag whether to append the file name and line number where the logger was called, by default True.
        """
        if append_filename_and_lineno:
            message = Logger.prepend_filename_and_lineno_to_message(message, **kwargs)
        self.debug_logger.debug(message)


    def info(
            self, 
            message: str, 
            append_filename_and_lineno: bool=True, 
            **kwargs
        ):
        """Log an info message.

        Parameters
        ----------
        message : str
            Logged message.
        append_filename_and_lineno : bool, optional
            Flag whether to append the file name and line number where the logger was called, by default True.
        """
        if append_filename_and_lineno:
            message = Logger.prepend_filename_and_lineno_to_message(message, **kwargs)
        self.debug_logger.info(message)
        self.info_logger.info(message)


    def warning(
            self, 
            message: str, 
            append_filename_and_lineno: bool=True, 
            **kwargs
        ):
        """Log a warning message.

        Parameters
        ----------
        message : str
            Logged message.
        append_filename_and_lineno : bool, optional
            Flag whether to append the file name and line number where the logger was called, by default True.
        """
        if append_filename_and_lineno:
            message = Logger.prepend_filename_and_lineno_to_message(message, **kwargs)
        self.debug_logger.warning(message)
        self.info_logger.warning(message)
        self.warning_logger.warning(message)


    def error(
            self, 
            message: str, 
            append_filename_and_lineno: bool=True, 
            **kwargs
        ):
        """Log an error message.

        Parameters
        ----------
        message : str
            Logged message.
        append_filename_and_lineno : bool, optional
            Flag whether to append the file name and line number where the logger was called, by default True.
        """
        if append_filename_and_lineno:
            message = Logger.prepend_filename_and_lineno_to_message(message, **kwargs)
        self.debug_logger.error(message)
        self.info_logger.error(message)
        self.warning_logger.error(message)
        self.error_logger.error(message)


    def critical(
            self, 
            message: str, 
            append_filename_and_lineno: bool=True, 
            **kwargs
        ):
        """Log a critical message.

        Parameters
        ----------
        message : str
            Logged message.
        append_filename_and_lineno : bool, optional
            Flag whether to append the file name and line number where the logger was called, by default True.
        """
        if append_filename_and_lineno:
            message = Logger.prepend_filename_and_lineno_to_message(message, **kwargs)
        self.debug_logger.critical(message)
        self.info_logger.critical(message)
        self.warning_logger.critical(message)
        self.error_logger.critical(message)
        self.critical_logger.critical(message)


    ### Pre-defined messaging ##########################################################################################

    # def specific_log_message(self, par1, par2):
    #     self._log(
    #         self.log_messages["specific-log-message"]["level"],
    #         self.log_messages["specific-log-message"]["message"].format(par1=par1, par2=par2)
    #     )


    ############################################################################
    ### Generic ################################################################
    ############################################################################

    def log_current_round(
        self,
        current_round: int
    ):
        self._log(
            self.log_messages["current_round"]["level"],
            self.log_messages["current_round"]["message"].format(current_round=current_round)
        )

    ############################################################################
    ### Daemon #################################################################
    ############################################################################

    def log_maintaining_valads(
        self,
        num_of_valads: int
    ):
        self._log(
            self.log_messages["maintaining_valads"]["level"],
            self.log_messages["maintaining_valads"]["message"].format(num_of_valads=num_of_valads)
        )

    def log_state_of_valad_with_id(
        self,
        app_id: int,
        state: bytes
    ):
        self._log(
            self.log_messages["state_of_valad_with_id"]["level"],
            self.log_messages["state_of_valad_with_id"]["message"].format(app_id=app_id, state=state)
        )

    def log_set_valad_ready_attribute_error(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["set_valad_ready_attribute_error"]["level"],
            self.log_messages["set_valad_ready_attribute_error"]["message"].format(app_id=app_id)
        )

    def log_maintaining_delcos(
        self,
        num_of_delcos: int
    ):
        self._log(
            self.log_messages["maintaining_delcos"]["level"],
            self.log_messages["maintaining_delcos"]["message"].format(num_of_delcos=num_of_delcos)
        )

    def log_unknown_delco_error(
        self,
        app_id: int,
        e: Exception
    ):
        self._log(
            self.log_messages["unknown_delco_error"]["level"],
            self.log_messages["unknown_delco_error"]["message"].format(app_id=app_id, e=e)
        )

    def log_removed_ended_or_deleted_delco(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["removed_ended_or_deleted_delco"]["level"],
            self.log_messages["removed_ended_or_deleted_delco"]["message"].format(app_id=app_id)
        )

    def log_state_of_delco_with_id(
        self,
        app_id: int,
        state: bytes
    ):
        self._log(
            self.log_messages["state_of_delco_with_id"]["level"],
            self.log_messages["state_of_delco_with_id"]["message"].format(app_id=app_id, state=state)
        )

    def log_unknown_delco_state(
        self,
        state: bytes
    ):
        self._log(
            self.log_messages["unknown_delco_state"]["level"],
            self.log_messages["unknown_delco_state"]["message"].format(state=state)
        )

    def log_delco_in_ready_handler(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["delco_in_ready_handler"]["level"],
            self.log_messages["delco_in_ready_handler"]["message"].format(app_id=app_id)
        )

    def log_urlerror_checking_partkey_generated(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["urlerror_checking_partkey_generated"]["level"],
            self.log_messages["urlerror_checking_partkey_generated"]["message"].format(app_id=app_id)
        )

    def log_partkeys_generated_for_delco(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["partkeys_generated_for_delco"]["level"],
            self.log_messages["partkeys_generated_for_delco"]["message"].format(app_id=app_id)
        )

    def log_delco_cannot_pay(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["delco_cannot_pay"]["level"],
            self.log_messages["delco_cannot_pay"]["message"].format(app_id=app_id)
        )

    def log_attributeerror_cannot_pay(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["attributeerror_cannot_pay"]["level"],
            self.log_messages["attributeerror_cannot_pay"]["message"].format(app_id=app_id)
        )

    def log_logicerror_cannot_pay(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["logicerror_cannot_pay"]["level"],
            self.log_messages["logicerror_cannot_pay"]["message"].format(app_id=app_id)
        )

    def log_httperror_cannot_pay(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["httperror_cannot_pay"]["level"],
            self.log_messages["httperror_cannot_pay"]["message"].format(app_id=app_id)
        )

    def log_partkeys_not_submitted(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["partkeys_not_submitted"]["level"],
            self.log_messages["partkeys_not_submitted"]["message"].format(app_id=app_id)
        )

    def log_attributeerror_partkeys_not_submitted(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["attributeerror_partkeys_not_submitted"]["level"],
            self.log_messages["attributeerror_partkeys_not_submitted"]["message"].format(app_id=app_id)
        )

    def log_logicerror_partkeys_not_submitted(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["logicerror_partkeys_not_submitted"]["level"],
            self.log_messages["logicerror_partkeys_not_submitted"]["message"].format(app_id=app_id)
        )

    def log_httperror_partkeys_not_submitted(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["httperror_partkeys_not_submitted"]["level"],
            self.log_messages["httperror_partkeys_not_submitted"]["message"].format(app_id=app_id)
        )

    def log_partkey_params_submitted(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["partkey_params_submitted"]["level"],
            self.log_messages["partkey_params_submitted"]["message"].format(app_id=app_id)
        )

    def log_attributeerror_partkey_submit(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["attributeerror_partkey_submit"]["level"],
            self.log_messages["attributeerror_partkey_submit"]["message"].format(app_id=app_id)
        )

    def log_urlerror_checking_partkey_pending(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["urlerror_checking_partkey_pending"]["level"],
            self.log_messages["urlerror_checking_partkey_pending"]["message"].format(app_id=app_id)
        )

    def log_partkey_generation_pending(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["partkey_generation_pending"]["level"],
            self.log_messages["partkey_generation_pending"]["message"].format(app_id=app_id)
        )

    def log_requested_partkey_generation(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["requested_partkey_generation"]["level"],
            self.log_messages["requested_partkey_generation"]["message"].format(app_id=app_id)
        )

    def log_partkey_generation_denied(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["partkey_generation_denied"]["level"],
            self.log_messages["partkey_generation_denied"]["message"].format(app_id=app_id)
        )

    def log_partkeys_not_confirmed(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["partkeys_not_confirmed"]["level"],
            self.log_messages["partkeys_not_confirmed"]["message"].format(app_id=app_id)
        )

    def log_attributeerror_partkeys_not_confirmed(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["attributeerror_partkeys_not_confirmed"]["level"],
            self.log_messages["attributeerror_partkeys_not_confirmed"]["message"].format(app_id=app_id)
        )

    def log_logicerror_partkeys_not_confirmed(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["logicerror_partkeys_not_confirmed"]["level"],
            self.log_messages["logicerror_partkeys_not_confirmed"]["message"].format(app_id=app_id)
        )

    def log_httperror_partkeys_not_confirmed(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["httperror_partkeys_not_confirmed"]["level"],
            self.log_messages["httperror_partkeys_not_confirmed"]["message"].format(app_id=app_id)
        )

    def log_delco_in_live_handler(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["delco_in_live_handler"]["level"],
            self.log_messages["delco_in_live_handler"]["message"].format(app_id=app_id)
        )

    def log_contract_expired(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["delco_in_live_handler"]["level"],
            self.log_messages["delco_in_live_handler"]["message"].format(app_id=app_id)
        )

    def log_expired_attribute_error(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["expired_attribute_error"]["level"],
            self.log_messages["expired_attribute_error"]["message"].format(app_id=app_id)
        )

    def log_tried_contract_expired(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["tried_contract_expired"]["level"],
            self.log_messages["tried_contract_expired"]["message"].format(app_id=app_id)
        )

    def log_httperror_contract_expired(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["httperror_contract_expired"]["level"],
            self.log_messages["httperror_contract_expired"]["message"].format(app_id=app_id)
        )

    def log_gating_or_stake_limit_breached(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["gating_or_stake_limit_breached"]["level"],
            self.log_messages["gating_or_stake_limit_breached"]["message"].format(app_id=app_id)
        )

    def log_expired_attribute_error(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["expired_attribute_error"]["level"],
            self.log_messages["expired_attribute_error"]["message"].format(app_id=app_id)
        )

    def log_gating_or_stake_limit_breached(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["gating_or_stake_limit_breached"]["level"],
            self.log_messages["gating_or_stake_limit_breached"]["message"].format(app_id=app_id)
        )

    def log_gating_or_stake_limit_breached_attribute_error(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["gating_or_stake_limit_breached_attribute_error"]["level"],
            self.log_messages["gating_or_stake_limit_breached_attribute_error"]["message"].format(app_id=app_id)
        )

    def log_logicerror_gating_or_stake_limit_breached(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["logicerror_gating_or_stake_limit_breached"]["level"],
            self.log_messages["logicerror_gating_or_stake_limit_breached"]["message"].format(app_id=app_id)
        )

    def log_httperror_gating_or_stake_limit_breached(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["httperror_gating_or_stake_limit_breached"]["level"],
            self.log_messages["httperror_gating_or_stake_limit_breached"]["message"].format(app_id=app_id)
        )

    def log_delco_in_ended_handler(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["delco_in_ended_handler"]["level"],
            self.log_messages["delco_in_ended_handler"]["message"].format(app_id=app_id)
        )

    def log_scheduled_partkey_deletion_for_ended_or_deleted(
        self,
        app_id: int,
        scheduled_deletion: int,
        round_end: int
    ):
        self._log(
            self.log_messages["scheduled_partkey_deletion_for_ended_or_deleted"]["level"],
            self.log_messages["scheduled_partkey_deletion_for_ended_or_deleted"]["message"].format(
                app_id=app_id,
                scheduled_deletion=scheduled_deletion,
                round_end=round_end
            )
        )

    def log_no_partkeys_found_for_ended_or_deleted(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["no_partkeys_found_for_ended_or_deleted"]["level"],
            self.log_messages["no_partkeys_found_for_ended_or_deleted"]["message"].format(app_id=app_id)
        )

    def log_delco_in_deleted_handler(
        self,
        app_id: int
    ):
        self._log(
            self.log_messages["delco_in_deleted_handler"]["level"],
            self.log_messages["delco_in_deleted_handler"]["message"].format(app_id=app_id)
        )

    def log_num_of_valad_ids_found(
        self,
        num_of_valads: int
    ):
        self._log(
            self.log_messages["num_of_valad_ids_found"]["level"],
            self.log_messages["num_of_valad_ids_found"]["message"].format(num_of_valads=num_of_valads)
        )

    def log_num_of_valad_clients_connected(
        self,
        num_of_valads: int
    ):
        self._log(
            self.log_messages["num_of_valad_clients_connected"]["level"],
            self.log_messages["num_of_valad_clients_connected"]["message"].format(num_of_valads=num_of_valads)
        )

    def log_num_of_updated_valads(
        self,
        num_of_updated_valads:int,
        num_of_valads: int
    ):
        self._log(
            self.log_messages["num_of_updated_valads"]["level"],
            self.log_messages["num_of_updated_valads"]["message"].format(
                num_of_updated_valads=num_of_updated_valads, 
                num_of_valads=num_of_valads
            )
        )

    def log_zero_valad_clients(
        self,
        valad_id_list: list
    ):
        self._log(
            self.log_messages["zero_valad_clients"]["level"],
            self.log_messages["zero_valad_clients"]["message"].format(valad_id_list=valad_id_list)
        )

    def log_zero_valad_clients(
        self,
        valad_id_list: list
    ):
        self._log(
            self.log_messages["zero_valad_clients"]["level"],
            self.log_messages["zero_valad_clients"]["message"].format(valad_id_list=valad_id_list)
        )

    def log_num_of_connected_delcos(
        self,
        num_of_delcos: int
    ):
        self._log(
            self.log_messages["num_of_connected_delcos"]["level"],
            self.log_messages["num_of_connected_delcos"]["message"].format(num_of_delcos=num_of_delcos)
        )

    def log_num_of_delco_clients_connected(
        self,
        num_of_delcos: int
    ):
        self._log(
            self.log_messages["num_of_delco_clients_connected"]["level"],
            self.log_messages["num_of_delco_clients_connected"]["message"].format(num_of_delcos=num_of_delcos)
        )

    def log_num_of_updated_delcos(
        self,
        num_of_updated_delcos:int,
        num_of_delcos: int
    ):
        self._log(
            self.log_messages["num_of_updated_delcos"]["level"],
            self.log_messages["num_of_updated_delcos"]["message"].format(
                num_of_updated_delcos=num_of_updated_delcos, 
                num_of_delcos=num_of_delcos
            )
        )

    def log_algod_ok_continuing(
        self
    ):
        self._log(
            self.log_messages["algod_ok_continuing"]["level"],
            self.log_messages["algod_ok_continuing"]["message"]
        )

    def log_generic_contract_servicing_error(
        self,
        e: Exception
    ):
        self._log(
            self.log_messages["generic_contract_servicing_error"]["level"],
            self.log_messages["generic_contract_servicing_error"]["message"].format(e=e)
        )

    def log_generic_partkey_manager_error(
        self,
        e: Exception
    ):
        self._log(
            self.log_messages["generic_partkey_manager_error"]["level"],
            self.log_messages["generic_partkey_manager_error"]["message"].format(e=e)
        )

    def log_algod_error(
        self,
        msg: int
    ):
        self._log(
            self.log_messages["algod_error"]["level"],
            self.log_messages["algod_error"]["message"].format(msg=msg)
        )

    def log_single_loop_execution_time(
        self,
        duration_s: float
    ):
        self._log(
            self.log_messages["single_loop_execution_time"]["level"],
            self.log_messages["single_loop_execution_time"]["message"].format(duration_s=duration_s)
        )

    def log_targeted_sleep_duration(
        self,
        duration_s: float
    ):
        self._log(
            self.log_messages["targeted_sleep_duration"]["level"],
            self.log_messages["targeted_sleep_duration"]["message"].format(duration_s=duration_s)
        )

    def log_could_not_sleep(
        self,
        duration_s: float,
        e: Exception
    ):
        self._log(
            self.log_messages["could_not_sleep"]["level"],
            self.log_messages["could_not_sleep"]["message"].format(duration_s=duration_s, e=e)
        )

    ############################################################################
    ### AppWrapper #############################################################
    ############################################################################

    def log_app_create_urlerror(
        self,
        app_id: int,
        errno: int,
        strerror=str
    ):
        self._log(
            self.log_messages["app_create_urlerror"]["level"],
            self.log_messages["app_create_urlerror"]["message"].format(
                app_id=app_id,
                errno=errno,
                strerror=strerror
            )
        )

    def log_app_create_algohttperror(
        self,
        app_id: int,
        errno: int,
        strerror=str
    ):
        self._log(
            self.log_messages["app_create_algohttperror"]["level"],
            self.log_messages["app_create_algohttperror"]["message"].format(
                app_id=app_id,
                errno=errno,
                strerror=strerror
            )
        )

    def log_app_create_genericerror(
        self,
        app_id: float,
        e: Exception
    ):
        self._log(
            self.log_messages["app_create_genericerror"]["level"],
            self.log_messages["app_create_genericerror"]["message"].format(app_id=app_id, e=e)
        )

    def log_app_dynamic_update_genericerror(
        self,
        app_id: float,
        e: Exception
    ):
        self._log(
            self.log_messages["app_dynamic_update_genericerror"]["level"],
            self.log_messages["app_dynamic_update_genericerror"]["message"].format(app_id=app_id, e=e)
        )

    ############################################################################
    ### Partkeymanager #########################################################
    ############################################################################
    
    def log_partkey_generation_request(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int
    ):
        self._log(
            self.log_messages["partkey_generation_request"]["level"],
            self.log_messages["partkey_generation_request"]["message"].format(
                address=address,
                vote_first_valid=vote_first_valid,
                vote_last_valid=vote_last_valid
            )
        )
    
    def log_generating_partkeys(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int
    ):
        self._log(
            self.log_messages["generating_partkeys"]["level"],
            self.log_messages["generating_partkeys"]["message"].format(
                address=address,
                vote_first_valid=vote_first_valid,
                vote_last_valid=vote_last_valid
            )
        )

    def log_requested_partkey_in_past(
        self,
        num_of_keys: int
    ):
        self._log(
            self.log_messages["requested_partkey_in_past"]["level"],
            self.log_messages["requested_partkey_in_past"]["message"].format(num_of_keys=num_of_keys)
        )

    def log_pending_buffer_is_full(
        self,
        num_of_keys: int
    ):
        self._log(
            self.log_messages["pending_buffer_is_full"]["level"],
            self.log_messages["pending_buffer_is_full"]["message"].format(num_of_keys=num_of_keys)
        )

    def log_generated_buffer_is_full(
        self,
        num_of_keys: int
    ):
        self._log(
            self.log_messages["generated_buffer_is_full"]["level"],
            self.log_messages["generated_buffer_is_full"]["message"].format(num_of_keys=num_of_keys)
        )

    def log_requested_partkey_in_pending(
        self
    ):
        self._log(
            self.log_messages["requested_partkey_in_pending"]["level"],
            self.log_messages["requested_partkey_in_pending"]["message"]
        )

    def log_requested_partkey_in_generated(
        self
    ):
        self._log(
            self.log_messages["requested_partkey_in_generated"]["level"],
            self.log_messages["requested_partkey_in_generated"]["message"]
        )

    def log_partkey_generation_request_added(
        self
    ):
        self._log(
            self.log_messages["partkey_generation_request_added"]["level"],
            self.log_messages["partkey_generation_request_added"]["message"]
        )

    def log_generic_algod_error(
        self,
        e: Exception
    ):
        self._log(
            self.log_messages["generic_algod_error"]["level"],
            self.log_messages["generic_algod_error"]["message"].format(e=e)
        )



"""Various utility functions, including interaction with the Valar Smart Contracts.
"""
import base64
import struct
import dataclasses

from algosdk.error import AlgodHTTPError
from algokit_utils.beta.algorand_client import AlgorandClient
from algokit_utils.beta.account_manager import AddressAndSigner
from algokit_utils import TransactionParameters, ABITransactionResponse
from algokit_utils.network_clients import AlgoClientConfig, AlgoClientConfigs
from algosdk.abi import AddressType, ArrayStaticType, ByteType, TupleType, UintType

from valar_daemon.NoticeboardClient import NoticeboardClient, KeyRegTxnInfo
from valar_daemon.DelegatorContractClient import (
    DelegationTermsBalance,
    DelegationTermsGeneral,
    GlobalState as DelcoGlobalState
)
from valar_daemon.ValidatorAdClient import ValidatorTermsTiming
from valar_daemon.constants import *



### Misc utilities #####################################################################################################

def get_algorand_client(
    algod_config_server: str,
    algod_config_token: str
) -> AlgorandClient:
    """Get an algorand client from the provided configuration.

    Parameters
    ----------
    algod_config_server : str
        Algod URL.
    algod_config_token : str
        Algod token.

    Returns
    -------
    AlgorandClient
        Configured Algorand client without timeout.
    """
    algod_config = AlgoClientConfig(
        server=algod_config_server,
        token=algod_config_token
    )
    indexer_config = AlgoClientConfig(
        server='http://localhost:8980',
        token='aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
    )
    kmd_config = AlgoClientConfig(
        server='http://localhost:4002',
        token='aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
    )
    algorand_client = AlgorandClient(
        AlgoClientConfigs(
            algod_config=algod_config,
            indexer_config=indexer_config,
            kmd_config=kmd_config,
        )
    )
    algorand_client.set_suggested_params_timeout(0)
    return algorand_client


def decode_uint64_list(
    data: bytes
) -> list:
    """Decode a string of bytes.

    Args:
        data (bytes): String of bytes.

    Returns:
        list: Decoded bytes.
    """
    # Determine the number of uint64 values in the bytes object
    num_uint64 = len(data) // 8

    # Unpack the bytes object into a list of uint64 values (big-endian)
    int_list = list(struct.unpack(f">{num_uint64}Q", data))

    return int_list


@dataclasses.dataclass()
class User:
    role: bytes
    address: str

    def to_box_name(self) -> bytes:
        return self.role + AddressType().encode(self.address)


# @dataclasses.dataclass()
# class SecurityDeposit:
#     blocked: int = 0
#     total: int = 0

#     @classmethod
#     def from_bytes(cls, data: bytes) -> "SecurityDeposit":
#         data_type = TupleType(
#             [
#                 UintType(64),   # blocked
#                 UintType(64),   # total
#             ]
#         )
#         decoded_tuple = data_type.decode(data)

#         return SecurityDeposit(
#             blocked=decoded_tuple[0],
#             total=decoded_tuple[1],
#         )



### Delegator contract client helpers ##################################################################################

from typing import Tuple, List
def get_delco_fee_and_gating_asa_id(
    global_state_raw: DelcoGlobalState
) -> Tuple[int, List[int]]:
    """Get the fee and gating ASA IDs by interpreting the delco global state.

    Notes
    -----
    Algo ASA is added by default, since it is staked.
    The payment ASA is also included in the gating ASA list.

    Parameters
    ----------
    global_state_raw : DelcoGlobalState
        Uninterpreted delco global state.

    Returns
    -------
    Tuple[int, List[int]]
    """
    global_state = DelegatorContractGlobalState.from_global_state(global_state_raw)
    gating_asa_id_list = [ALGO_ASA_ID] # By defaults, since ALGO is the one staked
    for asa in global_state.delegation_terms_balance.gating_asa_list:
        if asa[0] != ALGO_ASA_ID:
            gating_asa_id_list.append(asa[0]) # Add up to 4 additional ASAs
    return (
        global_state.delegation_terms_general.fee_asset_id,
        gating_asa_id_list
    )


DEFAULT_DELEGATION_TERMS_BALANCE = DelegationTermsBalance(
    stake_max = 0,
    cnt_breach_del_max = 0,
    rounds_breach = 0,
    gating_asa_list = [(0, 0), (0, 0)],
)


DEFAULT_DELEGATION_TERMS_GENERAL = DelegationTermsGeneral(
    commission = 0,
    fee_round = 0,
    fee_setup = 0,
    fee_asset_id = 0,
    partner_address = ZERO_ADDRESS,
    fee_round_partner = 0,
    fee_setup_partner = 0,
    rounds_setup = 0,
    rounds_confirm = 0,
)


def decode_abi_address(data: bytes) -> str:
    return AddressType().decode(data)


def decode_delegation_terms_balance(data: bytes) -> DelegationTermsBalance:
    delegation_terms_balance_type = TupleType(
        [
            UintType(64),  # stake_max
            UintType(64),  # cnt_breach_del_max
            UintType(64),  # rounds_breach
            ArrayStaticType(ArrayStaticType(UintType(64), 2), 2),  # gating_asa_list
        ]
    )

    decoded_tuple = delegation_terms_balance_type.decode(data)

    delegation_terms_balance = DelegationTermsBalance(
        stake_max=decoded_tuple[0],
        cnt_breach_del_max=decoded_tuple[1],
        rounds_breach=decoded_tuple[2],
        gating_asa_list=[tuple(item) for item in decoded_tuple[3]],
    )

    return delegation_terms_balance


def decode_delegation_terms_general(data: bytes) -> DelegationTermsGeneral:
    delegation_terms_general_type = TupleType(
        [
            UintType(64),  # commission
            UintType(64),  # fee_round
            UintType(64),  # fee_setup
            UintType(64),  # fee_asset_id
            AddressType(),  # partner_address
            UintType(64),  # fee_round_partner
            UintType(64),  # fee_setup_partner
            UintType(64),  # rounds_setup
            UintType(64),  # rounds_confirm
        ]
    )

    decoded_tuple = delegation_terms_general_type.decode(data)

    delegation_terms_general = DelegationTermsGeneral(
        commission=decoded_tuple[0],
        fee_round=decoded_tuple[1],
        fee_setup=decoded_tuple[2],
        fee_asset_id=decoded_tuple[3],
        partner_address=decoded_tuple[4],
        fee_round_partner=decoded_tuple[5],
        fee_setup_partner=decoded_tuple[6],
        rounds_setup=decoded_tuple[7],
        rounds_confirm=decoded_tuple[8],
    )

    return delegation_terms_general


@dataclasses.dataclass(kw_only=True)
class DelegatorContractGlobalState:
    cnt_breach_del: int = 0
    del_beneficiary: str = ZERO_ADDRESS
    del_manager: str = ZERO_ADDRESS
    delegation_terms_balance: DelegationTermsBalance = dataclasses.field(default_factory=lambda: DEFAULT_DELEGATION_TERMS_BALANCE)  # noqa: E501
    delegation_terms_general: DelegationTermsGeneral = dataclasses.field(default_factory=lambda: DEFAULT_DELEGATION_TERMS_GENERAL)  # noqa: E501
    fee_operational: int = 0
    fee_operational_partner: int = 0
    noticeboard_app_id: int = 0
    round_breach_last: int = 0
    round_claim_last: int = 0
    round_end: int = 0
    round_ended: int = 0
    round_expiry_soon_last: int = 0
    round_start: int = 0
    sel_key: bytes = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"  # noqa: E501
    state: bytes = b"\x00"
    state_proof_key: bytes = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"  # noqa: E501
    tc_sha256: bytes = bytes(32)
    validator_ad_app_id: int = 0
    vote_key: bytes = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"  # noqa: E501
    vote_key_dilution: int = 0

    @classmethod
    def from_global_state(cls, gs: DelcoGlobalState) -> "DelegatorContractGlobalState":
        return cls(
            cnt_breach_del = gs.cnt_breach_del,
            del_beneficiary = decode_abi_address(gs.del_beneficiary.as_bytes),
            del_manager = decode_abi_address(gs.del_manager.as_bytes),
            delegation_terms_balance = decode_delegation_terms_balance(gs.delegation_terms_balance.as_bytes),
            delegation_terms_general = decode_delegation_terms_general(gs.delegation_terms_general.as_bytes),
            fee_operational = gs.fee_operational,
            fee_operational_partner = gs.fee_operational_partner,
            noticeboard_app_id = gs.noticeboard_app_id,
            round_breach_last = gs.round_breach_last,
            round_claim_last = gs.round_claim_last,
            round_end = gs.round_end,
            round_ended = gs.round_ended,
            round_expiry_soon_last = gs.round_expiry_soon_last,
            round_start = gs.round_start,
            sel_key = gs.sel_key.as_bytes,
            state = gs.state.as_bytes,
            state_proof_key = gs.state_proof_key.as_bytes,
            tc_sha256 = gs.tc_sha256.as_bytes,
            validator_ad_app_id = gs.validator_ad_app_id,
            vote_key = gs.vote_key.as_bytes,
            vote_key_dilution = gs.vote_key_dilution,
        )

    @classmethod
    def with_defaults(cls) -> "DelegatorContractGlobalState":
        return cls()



### Validator ad client helpers ########################################################################################

DEFAULT_VALIDATOR_TERMS_TIME = ValidatorTermsTiming(
    rounds_setup = 0,
    rounds_confirm = 0,
    rounds_duration_min = 0,
    rounds_duration_max = 0,
    round_max_end = 0,
)


def decode_validator_terms_time(data: bytes) -> ValidatorTermsTiming:
    data_type = TupleType(
        [
            UintType(64),  # rounds_setup
            UintType(64),  # rounds_confirm
            UintType(64),  # rounds_duration_min
            UintType(64),  # rounds_duration_max
            UintType(64),  # round_max_end
        ]
    )

    decoded_tuple = data_type.decode(data)

    decoded_data = ValidatorTermsTiming(
        rounds_setup = decoded_tuple[0],
        rounds_confirm = decoded_tuple[1],
        rounds_duration_min = decoded_tuple[2],
        rounds_duration_max = decoded_tuple[3],
        round_max_end = decoded_tuple[4],
    )

    return decoded_data



### Valar Smart Contract interface #####################################################################################


@dataclasses.dataclass()
class UserInfo:
    """
    UserInfo(): Creates a new object with user info with all values initialized to zero by default.
    """
    role: bytes = b""
    dll_name: bytes = b""

    prev_user: str = ZERO_ADDRESS
    next_user: str = ZERO_ADDRESS

    app_ids: list[int] = dataclasses.field(default_factory=lambda: [0] * 110)
    cnt_app_ids: int = 0

    @classmethod
    def from_bytes(cls, data: bytes) -> "UserInfo":
        data_type = TupleType(
            [
                ArrayStaticType(ByteType(), 4),    # role
                ArrayStaticType(ByteType(), 8),    # dll_name
                AddressType(),   # prev_user
                AddressType(),   # next_user
                ArrayStaticType(UintType(64), 110),   # app_ids
                UintType(64),   # cnt_app_ids
            ]
        )
        decoded_tuple = data_type.decode(data)

        return UserInfo(
            role=bytes(decoded_tuple[0]),
            dll_name=bytes(decoded_tuple[1]),
            prev_user=decoded_tuple[2],
            next_user=decoded_tuple[3],
            app_ids= decoded_tuple[4],
            cnt_app_ids=decoded_tuple[5],
        )

    def get_free_app_idx(self) -> int | None:
        return self.get_app_idx(0)

    def get_app_idx(self, app_id: int) -> int | None:
        try:
            val_app_idx = self.app_ids.index(app_id)
        except ValueError:
            val_app_idx = None
        return val_app_idx
    

def app_get_user_info(
    algorand_client,
    noticeboard_app_id,
    user: str,
) -> UserInfo | None:
    box_raw = app_user_box(
        algorand_client,
        noticeboard_app_id,
        user
    )
    if box_raw[1]:
        user_info = UserInfo.from_bytes(box_raw[0])
    else:
        user_info = None
    return user_info


def app_user_box(
    algorand_client,
    noticeboard_app_id,
    user: str,
) -> tuple[bytes, bool]:
    box_name = AddressType().encode(user)
    box_raw = app_box(
        algorand_client,
        noticeboard_app_id,
        box_name=box_name
    )
    return box_raw


def app_box(
    algorand_client,
    noticeboard_app_id,
    box_name: bytes,
) -> tuple[bytes, bool]:
    return get_box(
        algorand_client=algorand_client,
        box_name=box_name,
        app_id=noticeboard_app_id
    )


def get_box(
    algorand_client : AlgorandClient,
    box_name : bytes,
    app_id : int,
) -> tuple[bytes, bool]:

    exists = False
    box_raw = b""

    try:
        box_raw = algorand_client.client.algod.application_box_by_name(
            application_id=app_id,
            box_name=box_name,
        )
        box_raw = base64.b64decode(box_raw["value"])
        exists = True
    except AlgodHTTPError:
        pass
    except Exception as e:
        print(e)

    return [box_raw, exists]


def get_val_and_del_app_idx(
    algorand_client: AlgorandClient,
    noticeboard_app_id: int,
    valown_address: str,
    delman_address: str,
    valad_id: int,
    delco_id: int
) -> Tuple[int, int]:
    """Get the app indexes at which valad and delco are stored for the corresponding valown and delman.

    Parameters
    ----------
    algorand_client : AlgorandClient
        Algorand client.
    noticeboard_app_id : int
        Noticeboard ID.
    valown_address : str
        Validator owner address.
    delman_address : str
        Delegator manager address.
    valad_id : int
        Validator ad ID.
    delco_id : int
        Delegator contract ID.

    Returns
    -------
    Tuple[int, int]
        Valad and delco indexes.
    """
    # Get index in validator's list at which the validator ad is stored
    valown_user_info = app_get_user_info(
        algorand_client,
        noticeboard_app_id,
        valown_address
    )
    val_app_idx = valown_user_info.get_app_idx(valad_id)
    # Get index in delegator's list at which the delegator contract is stored
    delman_user_info = app_get_user_info(
        algorand_client,
        noticeboard_app_id,
        delman_address
    )
    del_app_idx = delman_user_info.get_app_idx(delco_id)
    return (val_app_idx, del_app_idx)


def valown_and_delman_boxes(
    valown_address: str,
    delman_address: str
) -> List[Tuple[int, bytes]]:
    """Prepare for box fetching from the noticeboard for the validator owner and delegator manager.

    Parameters
    ----------
    valown_address : str
        Validator owner address.
    delman_address : str
        Delegator manager address.

    Returns
    -------
    List[Tuple[int, bytes]]
        Valown and delman box.
    """
    return[
        (0,  AddressType().encode(valown_address)),
        (0,  AddressType().encode(delman_address))
    ]


def set_valad_ready(
    algorand_client: AlgorandClient,
    valown_address: str,
    valad_id: int,
    valman: AddressAndSigner,
    noticeboard_client: NoticeboardClient,
) -> ABITransactionResponse[None]:
    """Set validator ad to ready.

    Parameters
    ----------
    algorand_client : AlgorandClient
        Algorand client.
    valown_address : str
        Validator ad owner address.
    delman_address : str
        Delegator manager address.
    valad_id : int
        Validator ad app id.
    delco_id : int
        Delegator contract app id.
    fee_asset_id : int
        ID of the asset user for payments.
    valman : AddressAndSigner
        Validator manager address and signer.
    noticeboard_client : NoticeboardClient
        Noticeboard client.
    """

    boxes_user = [
        (0,  AddressType().encode(valown_address)),
    ]

    # Get index of app in user's list
    user_info = app_get_user_info(
        algorand_client,
        noticeboard_client.app_id,
        valown_address
    )
    val_app_idx = user_info.get_app_idx(valad_id)

    # Increase fee for inner txns for ad_ready.
    sp = algorand_client.client.algod.suggested_params()
    sp.fee = 2 * sp.min_fee
    sp.flat_fee = True

    return noticeboard_client.ad_ready(
        val_owner=valown_address,
        val_app=valad_id,
        val_app_idx=val_app_idx,
        ready=True,
        transaction_parameters = TransactionParameters(
            sender = valman.address,
            signer = valman.signer,
            suggested_params=sp,
            boxes=boxes_user
        ),
    )


def report_delco_breach_pay(
    algorand_client: AlgorandClient,
    valown_address: str,
    delman_address: str,
    valad_id: int,
    delco_id: int,
    fee_asset_id: int,
    valman: AddressAndSigner,
    noticeboard_client: NoticeboardClient
) -> ABITransactionResponse[None]:
    """Report that the delegator contract can not transfer payment funds to validator ad.

    Parameters
    ----------
    algorand_client : AlgorandClient
        Algorand client.
    valown_address : str
        Validator ad owner address.
    delman_address : str
        Delegator manager address.
    valad_id : int
        Validator ad app id.
    delco_id : int
        Delegator contract app id.
    fee_asset_id : int
        ID of the asset user for payments.
    valman : AddressAndSigner
        Validator manager address and signer.
    noticeboard_client : NoticeboardClient
        Noticeboard client.

    Returns
    -------
    ABITransactionResponse[None]
    """
    boxes = valown_and_delman_boxes(
        valown_address=valown_address,
        delman_address=delman_address
    )

    val_app_idx, del_app_idx = get_val_and_del_app_idx(
        algorand_client=algorand_client,
        noticeboard_app_id=noticeboard_client.app_id,
        valown_address=valown_address,
        delman_address=delman_address,
        valad_id=valad_id,
        delco_id=delco_id
    )

    # Add asset to the foreign asset array and box array
    if fee_asset_id != ALGO_ASA_ID:
        foreign_assets = [fee_asset_id]
        boxes_asset = [(valad_id, BOX_ASA_KEY_PREFIX + fee_asset_id.to_bytes(8, byteorder="big"))]
        boxes += boxes_asset
    else:
        foreign_assets = None

    # Increase fee for forwarding the call to validator ad (1) and then delegator contract (1),
    # as well as for (potential) distribution of earnings (2), (potential) payout of partner fee (1),
    # and (potential) notification message (1).
    sp = algorand_client.client.algod.suggested_params()
    sp.fee = 7 * sp.min_fee
    sp.flat_fee = True

    return noticeboard_client.breach_pay(
        del_manager=delman_address,
        del_app=delco_id,
        del_app_idx=del_app_idx,
        val_owner=valown_address,
        val_app=valad_id,
        val_app_idx=val_app_idx,
        transaction_parameters = TransactionParameters(
            sender=valman.address,
            signer=valman.signer,
            suggested_params=sp,
            foreign_assets=foreign_assets,
            accounts=[delman_address],
            boxes=boxes
        ),
    )


def submit_partkeys(
    algorand_client: AlgorandClient,
    valown_address: str,
    delman_address: str,
    delben_address: str,
    partner_address: str,
    valad_id: int,
    delco_id: int,
    fee_asset_id: int,
    valman: AddressAndSigner,
    noticeboard_client: NoticeboardClient,
    vote_first: int,
    vote_last: int,
    vote_key_dilution: int,
    vote_pk: str,
    selection_pk: str,
    state_proof_pk: str
) -> ABITransactionResponse[None]:
    """Submit participation keys.

    Parameters
    ----------
    algorand_client : AlgorandClient
        Algorand client.
    valown_address : str
        Validator ad owner address.
    delman_address : str
        Delegator manager address.
    delben_address : str
        Delegator beneficiary address.
    partner_address : str
        Address of partner for commissions.
    valad_id : int
        Validator ad app id.
    delco_id : int
        Delegator contract app id.
    fee_asset_id : int
        ID of the asset user for payments.
    valman : AddressAndSigner
        Validator manager address and signer.
    noticeboard_client : NoticeboardClient
        Noticeboard client.
    vote_first : int
        First round when keys are valid.
    vote_last : int
        Last round when keys are valid.
    vote_key_dilution : int
        Vote key dilution.
    vote_pk : str
        Vote parameter of participation keys.
    selection_pk : str
        Selection parameter of participation keys.
    state_proof_pk : str
        State proof parameter of participation keys.

    Returns
    -------
    ABITransactionResponse[None]
    """
    boxes = valown_and_delman_boxes(
        valown_address=valown_address,
        delman_address=delman_address
    )

    val_app_idx, del_app_idx = get_val_and_del_app_idx(
        algorand_client=algorand_client,
        noticeboard_app_id=noticeboard_client.app_id,
        valown_address=valown_address,
        delman_address=delman_address,
        valad_id=valad_id,
        delco_id=delco_id
    )

    # Add asset to the foreign asset array and box array
    if fee_asset_id != ALGO_ASA_ID:
        foreign_assets = [fee_asset_id]
        boxes_asset = [(valad_id, BOX_ASA_KEY_PREFIX + fee_asset_id.to_bytes(8, byteorder="big"))]
        boxes += boxes_asset
    else:
        foreign_assets = None

    # Add foreign addresses (add manager even if equal to beneficiary)
    foreign_accounts = [partner_address, delman_address]

    # Increase fee for forwarding the call to validator ad (1) and then delegator contract (1),
    # as well as distribution of earnings to the validator ad (1) and noticeboard (1),
    # (potential) payout of partner fee (1), and (potential) notification message (1).
    sp = algorand_client.client.algod.suggested_params()
    sp.fee = 7 * sp.min_fee
    sp.flat_fee = True

    return noticeboard_client.keys_submit(
        del_manager=delman_address,
        del_app=delco_id,
        del_app_idx=del_app_idx,
        val_owner=valown_address,
        val_app=valad_id,
        val_app_idx=val_app_idx,
        key_reg_txn_info = KeyRegTxnInfo(
            vote_first=vote_first,
            vote_last=vote_last,
            vote_key_dilution=vote_key_dilution,
            vote_pk=base64.b64decode(vote_pk),
            selection_pk=base64.b64decode(selection_pk),
            state_proof_pk=base64.b64decode(state_proof_pk),
            sender=delben_address,
        ),
        transaction_parameters = TransactionParameters(
            sender=valman.address,
            signer=valman.signer,
            suggested_params=sp,
            foreign_assets=foreign_assets,
            accounts=foreign_accounts,
            boxes=boxes,
        )
    )


def report_partkeys_not_submitted(
    algorand_client: AlgorandClient,
    valown_address: str,
    delman_address: str,
    # partner_address: str,
    valad_id: int,
    delco_id: int,
    fee_asset_id: int,
    valman: AddressAndSigner,
    noticeboard_client: NoticeboardClient
) -> ABITransactionResponse[None]:
    """Report that the validator manager has not submitted the partkeys on time.

    Parameters
    ----------
    algorand_client : AlgorandClient
        Algorand client.
    valown_address : str
        Validator ad owner address.
    delman_address : str
        Delegator manager address.
    # partner_address : str
    #     Address of partner for commissions.
    valad_id : int
        Validator ad app id.
    delco_id : int
        Delegator contract app id.
    fee_asset_id : int
        ID of the asset user for payments.
    valman : AddressAndSigner
        Validator manager address and signer.
    noticeboard_client : NoticeboardClient
        Noticeboard client.

    Returns
    -------
    ABITransactionResponse[None]
    """    
    boxes = valown_and_delman_boxes(
        valown_address=valown_address,
        delman_address=delman_address
    )

    val_app_idx, del_app_idx = get_val_and_del_app_idx(
        algorand_client=algorand_client,
        noticeboard_app_id=noticeboard_client.app_id,
        valown_address=valown_address,
        delman_address=delman_address,
        valad_id=valad_id,
        delco_id=delco_id
    )

    # Add asset to the foreign asset array and box array
    if fee_asset_id != ALGO_ASA_ID:
        foreign_assets = [fee_asset_id]
    else:
        foreign_assets = None

    # Increase fee for forwarding the call to validator ad (1) and then delegator contract (1),
    # plus the try for return of deposit to the delegator.
    sp = algorand_client.client.algod.suggested_params()
    sp.fee = 5 * sp.min_fee
    sp.flat_fee = True

    # foreign_accounts = [partner_address, delman_address]
    foreign_accounts = [delman_address]

    return noticeboard_client.keys_not_submitted(
        del_manager=delman_address,
        del_app=delco_id,
        del_app_idx=del_app_idx,
        val_owner=valown_address,
        val_app=valad_id,
        val_app_idx=val_app_idx,
        transaction_parameters = TransactionParameters(
            sender = valman.address,
            signer = valman.signer,
            suggested_params=sp,
            foreign_assets=foreign_assets,
            accounts=foreign_accounts,
            boxes=boxes,
        ),
    )


def report_unconfirmed_partkeys(
    algorand_client: AlgorandClient,
    valown_address: str,
    delman_address: str,
    partner_address: str,
    valad_id: int,
    delco_id: int,
    fee_asset_id: int,
    valman: AddressAndSigner,
    noticeboard_client: NoticeboardClient
) -> ABITransactionResponse[None]:
    """Report that the delegator beneficiary has not confirmed the partkeys on time.

    Parameters
    ----------
    algorand_client : AlgorandClient
        Algorand client.
    valown_address : str
        Validator ad owner address.
    delman_address : str
        Delegator manager address.
    partner_address : str
        Address of partner for commissions.
    valad_id : int
        Validator ad app id.
    delco_id : int
        Delegator contract app id.
    fee_asset_id : int
        ID of the asset user for payments.
    valman : AddressAndSigner
        Validator manager address and signer.
    noticeboard_client : NoticeboardClient
        Noticeboard client.

    Returns
    -------
    ABITransactionResponse[None]
    """    
    boxes = valown_and_delman_boxes(
        valown_address=valown_address,
        delman_address=delman_address
    )

    val_app_idx, del_app_idx = get_val_and_del_app_idx(
        algorand_client=algorand_client,
        noticeboard_app_id=noticeboard_client.app_id,
        valown_address=valown_address,
        delman_address=delman_address,
        valad_id=valad_id,
        delco_id=delco_id
    )

    # Add asset to the foreign asset array and box array
    if fee_asset_id != ALGO_ASA_ID:
        foreign_assets = [fee_asset_id]
    else:
        foreign_assets = None

    # Increase fee for forwarding the call to validator ad (1) and then delegator contract (1),
    # plus the try for return of deposit to the delegator.
    sp = algorand_client.client.algod.suggested_params()
    sp.fee = 6 * sp.min_fee
    sp.flat_fee = True

    foreign_accounts = [partner_address, delman_address]

    return noticeboard_client.keys_not_confirmed(
        del_manager=delman_address,
        del_app=delco_id,
        del_app_idx=del_app_idx,
        val_owner=valown_address,
        val_app=valad_id,
        val_app_idx=val_app_idx,
        transaction_parameters = TransactionParameters(
            sender = valman.address,
            signer = valman.signer,
            suggested_params=sp,
            accounts=foreign_accounts,
            foreign_assets=foreign_assets,
            boxes=boxes
        ),
    )


def report_contract_expired(
    algorand_client: AlgorandClient,
    valown_address: str,
    delman_address: str,
    partner_address: str,
    valad_id: int,
    delco_id: int,
    fee_asset_id: int,
    valman: AddressAndSigner,
    noticeboard_client: NoticeboardClient
) -> ABITransactionResponse[None]:
    """Report that the delegator contract has expired (ended due to completion).

    Parameters
    ----------
    algorand_client : AlgorandClient
        Algorand client.
    valown_address : str
        Validator ad owner address.
    delman_address : str
        Delegator manager address.
    partner_address : str
        Address of partner for commissions.
    valad_id : int
        Validator ad app id.
    delco_id : int
        Delegator contract app id.
    fee_asset_id : int
        ID of the asset user for payments.
    valman : AddressAndSigner
        Validator manager address and signer.
    noticeboard_client : NoticeboardClient
        Noticeboard client.

    Returns
    -------
    ABITransactionResponse[None]
    """    
    boxes = valown_and_delman_boxes(
        valown_address=valown_address,
        delman_address=delman_address
    )

    val_app_idx, del_app_idx = get_val_and_del_app_idx(
        algorand_client=algorand_client,
        noticeboard_app_id=noticeboard_client.app_id,
        valown_address=valown_address,
        delman_address=delman_address,
        valad_id=valad_id,
        delco_id=delco_id
    )

    # Add asset to the foreign asset array and box array
    if fee_asset_id != ALGO_ASA_ID:
        foreign_assets = [fee_asset_id]
        boxes_asset = [(valad_id, BOX_ASA_KEY_PREFIX + fee_asset_id.to_bytes(8, byteorder="big"))]
        boxes += boxes_asset
    else:
        foreign_assets = None

    # Add partners and delegator manager to foreign addresses
    foreign_accounts = [partner_address, delman_address]

    # Increase fee for forwarding the call to validator ad (1) and then delegator contract (1),
    # as well as for (potential) distribution of earnings (2), (potential) payout of partner fee (1),
    # and (potential) notification message (1).
    sp = algorand_client.client.algod.suggested_params()
    sp.fee = 7 * sp.min_fee
    sp.flat_fee = True

    return noticeboard_client.contract_expired(
        del_manager=delman_address,
        del_app=delco_id,
        del_app_idx=del_app_idx,
        val_owner=valown_address,
        val_app=valad_id,
        val_app_idx=val_app_idx,
        transaction_parameters = TransactionParameters(
            sender=valman.address,
            signer=valman.signer,
            suggested_params=sp,
            foreign_assets=foreign_assets,
            accounts=foreign_accounts,
            boxes=boxes,
        ),
    )


def report_delben_breach_limits(
    algorand_client: AlgorandClient,
    valown_address: str,
    delman_address: str,
    delben_address: str,
    partner_address: str,
    valad_id: int,
    delco_id: int,
    fee_asset_id: int,
    gating_asa_id_list: List[int],
    valman: AddressAndSigner,
    noticeboard_client: NoticeboardClient
) -> ABITransactionResponse[None]:
    """Report that the delegator beneficiary breached the max stake or one of the possible min gating ASA limits.

    Parameters
    ----------
    algorand_client : AlgorandClient
        Algorand client.
    valown_address : str
        Validator ad owner address.
    delman_address : str
        Delegator manager address.
    delben_address : str
        Delegator beneficiary address.
    partner_address : str
        Address of partner for commissions.
    valad_id : int
        Validator ad app id.
    delco_id : int
        Delegator contract app id.
    fee_asset_id : int
        ID of the asset user for payments.
    gating_asa_id_list : int
        IDs of assets (ASAs) usdeg for gating.
    valman : AddressAndSigner
        Validator manager address and signer.
    noticeboard_client : NoticeboardClient
        Noticeboard client.

    Returns
    -------
    ABITransactionResponse[None]
    """        
    boxes = valown_and_delman_boxes(
        valown_address=valown_address,
        delman_address=delman_address
    )

    val_app_idx, del_app_idx = get_val_and_del_app_idx(
        algorand_client=algorand_client,
        noticeboard_app_id=noticeboard_client.app_id,
        valown_address=valown_address,
        delman_address=delman_address,
        valad_id=valad_id,
        delco_id=delco_id
    )

    foreign_assets = [asa_id for asa_id in gating_asa_id_list if asa_id != ALGO_ASA_ID]
    
    if fee_asset_id != ALGO_ASA_ID:
        foreign_assets.append(fee_asset_id)

    if fee_asset_id != ALGO_ASA_ID:
        boxes_asset = [(valad_id, BOX_ASA_KEY_PREFIX + fee_asset_id.to_bytes(8, byteorder="big"))]
        boxes += boxes_asset

    # Add delegator manager account and delegator beneficiary account to the foreign account array
    foreign_accounts = [delben_address, delman_address, partner_address]

    # Increase fee for forwarding the call to validator ad (1) and then delegator contract (1),
    # as well as for (potential) distribution of earnings (2), (potential) note sending (1), and
    # the gas app call (1), the actual app call (1), (potential) payout of partner fee (1),
    # and (potential) notification message (1).
    sp = algorand_client.client.algod.suggested_params()
    sp.fee = 9 * sp.min_fee
    sp.flat_fee = True

    return noticeboard_client.compose(
    ).gas(
        transaction_parameters=TransactionParameters(
            sender = valman.address,
            signer = valman.signer,
            foreign_assets=foreign_assets,
            accounts=foreign_accounts,
            foreign_apps=[valad_id, delco_id]
        ),
    ).breach_limits(
        del_manager=delman_address,
        del_app=delco_id,
        del_app_idx=del_app_idx,
        val_owner=valown_address,
        val_app=valad_id,
        val_app_idx=val_app_idx,
        transaction_parameters = TransactionParameters(
            sender = valman.address,
            signer = valman.signer,
            suggested_params=sp,
            boxes=boxes,
        ),
    ).execute()



"""Valar Daemon.

Some terms are shortened to preserve space:
- valad: Validator Ad
- delco: Delegator Contract
- valman: Validator Manager.
"""
import copy
import time
from pathlib import Path
from dataclasses import dataclass
from urllib.error import URLError, HTTPError

from algosdk import mnemonic, account
from algokit_utils.beta.algorand_client import AlgorandClient
from algokit_utils.beta.account_manager import AddressAndSigner
from algosdk.atomic_transaction_composer import AccountTransactionSigner      
from algokit_utils.logic_error import LogicError
from algosdk.error import AlgodHTTPError

from valar_daemon.AppWrapper import ValadAppWrapperList, DelcoAppWrapperList, ValadAppWrapper, DelcoAppWrapper
from valar_daemon.DaemonConfig import DaemonConfig
from valar_daemon.PartkeyManager import PartkeyManager, PARTKEY_GENERATION_REQUEST_OK_ADDED
from valar_daemon.Logger import Logger
from valar_daemon.utils import (
    get_algorand_client,
    set_valad_ready, 
    submit_partkeys,
    report_delco_breach_pay,
    report_partkeys_not_submitted,
    report_unconfirmed_partkeys,
    report_contract_expired,
    report_delben_breach_limits,
)
from valar_daemon.constants import (
    VALAD_NOT_READY_STATUS_CHANGE_OK,
    VALAD_NOT_READY_STATUS_ATTRIBUTE_ERROR,
    VALAD_NOT_READY_STATUS_NO_CHANGE,
    DELCO_READY_STATUS_REQUEST_DENIED,
    DELCO_READY_STATUS_BREACH_PAY,
    DELCO_READY_STATUS_NOT_SUBMITTED,
    DELCO_READY_STATUS_URL_ERROR,
    DELCO_READY_STATUS_REQUESTED,
    DELCO_READY_STATUS_PENDING,
    DELCO_READY_STATUS_SUBMITTED,
    VALAD_STATE_NOT_READY,
    DELCO_STATE_READY,
    DELCO_STATE_SUBMITTED,
    DELCO_STATE_LIVE,
    DELCO_STATE_ENDED_WITHDREW
) 


@dataclass(kw_only=True)
class AlgodStatus:
    """Algod status abstraction class.

    Parameters
    ----------
    is_ok : bool
        Flag indicating whether algod is running OK.
    message : str
        Corresponding (error) message, obtained when checking the status.
    """
    is_ok: bool
    message: str


class Daemon(object):


    def __init__(
        self,
        log_path: str,
        config_path: str
    ):
        """Valar Daemon master class.

        Parameters
        ----------
        log_path : str
            Path to where the log will be generated.
        config_path : str
            Path to the input config file.
        daemon_config : DaemonConfig
            Daemon configuration.
        logger : Logger
            Daemon log abstraction.
        algorand_client : AlgorandClient
            Algorand client.
        partkey_manager : PartkeyManager
            Participation key management class.
        valman : AddressAndSigner
            Validator Manager.
        stop_flag : bool
            Indicator to stop the Daemon.
        loop_period_s : float
            Period at which contracts are checked.
        valad_app_list : ValadAppWrapperList
            List of Validator Ads and relevant info.
        delco_app_list : DelcoAppWrapperList
            List of Delegator Contract and relevant info.
        """
        ### Read config ################################################################################################
        config_filename = Path(config_path).name
        self.daemon_config = DaemonConfig(
            Path(config_path).parent,
            config_filename
        )
        self.daemon_config.read_config() # Read and load up config values.
        self.daemon_config.create_swap() # In case of intermediate config reading, allow the user to edit the original.

        ### Initialize logger ##########################################################################################
        self.logger = Logger(
            log_dirpath=log_path,
            log_max_size_bytes=self.daemon_config.max_log_file_size_B,
            log_file_count=self.daemon_config.num_of_log_files_per_level
        )

        ### Configure client ###########################################################################################
        self.algorand_client = get_algorand_client(
            algod_config_server = self.daemon_config.algod_config_server,
            algod_config_token = self.daemon_config.algod_config_token,
        )

        ### Set up partkey manager #####################################################################################
        self.partkey_manager = PartkeyManager(
            self.logger,
            self.algorand_client
        )

        ### Set up validator manager ###################################################################################
        manager_private_key = mnemonic.to_private_key(self.daemon_config.validator_manager_mnemonic)
        manager_address = account.address_from_private_key(manager_private_key)
        self.valman = AddressAndSigner(
            address=manager_address,
            signer=AccountTransactionSigner(manager_private_key)
        )

        ### Set up loop ################################################################################################
        self.stop_flag = False
        self.loop_period_s = self.daemon_config.loop_period_s

        ### Initialize up valad and delco lists ########################################################################
        self.valad_app_list = ValadAppWrapperList(
            self.algorand_client,
            self.logger
        )
        self.delco_app_list = DelcoAppWrapperList(
            self.algorand_client,
            self.logger
        )
        

    @staticmethod
    def check_algod_status(
        algorand_client: AlgorandClient
    ) -> AlgodStatus:
        """Check the status of algod and report error if not OK.

        Notes
        -----
        URLError 111: Algod service is offline
        URLError -2: Can not reach serice (internet offline or initialization URL/token is not correct)

        Parameters
        ----------
        algorand_client : AlgorandClient
            Algorand client.

        Returns
        -------
        AlgodStatus
        """
        try:
            algorand_client.client.algod.status()
            return AlgodStatus(
                is_ok=True, 
                message="Algod is OK."
            )
        except URLError as e:
            return AlgodStatus(
                is_ok = False,
                message = f'URLError {e.args[0].errno}: {e.args[0].strerror}'
            )
        except AlgodHTTPError as e:
            return AlgodStatus(
                is_ok = False,
                message = f'AlgodHTTPError {e.code}: {e.args[0]}'
            )
        except Exception as e:
            return AlgodStatus(
                is_ok = False,
                message = f'Unknown error: {e}'
            )


    def maintain_valads(
        self
    ):
        """Maintain validator ads (try to change state from `NOT READY` to `READY` where applicable).
        """
        self.logger.log_maintaining_valads(num_of_valads=len(self.valad_app_list.get_app_list()))
        for valad_app in self.valad_app_list.get_app_list():
            self.maintain_single_valad(
                self.algorand_client,
                self.valman,
                copy.copy(valad_app),
                self.logger
            )


    @staticmethod
    def maintain_single_valad(
        algorand_client: AlgorandClient,
        valman: AddressAndSigner,
        valad_app: ValadAppWrapper,
        logger: Logger,
    ) -> int:
        """Maintain validator ads (try to change state from `NOT READY` to `READY` where applicable).

        Parameters
        ----------
        algorand_client : AlgorandClient
            Algorand client.
        valman : AddressAndSigner
            Validator manager.
        valad_app : ValadAppWrapper
            Client and relevant parameters of the validator ad.
        logger : logging.Logger
            Message logger.

        Returns
        -------
        int
            Status message.
        """
        logger.log_state_of_valad_with_id(app_id=valad_app.app_id, state=valad_app.state)
        if valad_app.state == VALAD_STATE_NOT_READY:
            try:
                set_valad_ready(
                    algorand_client=algorand_client,
                    valown_address=valad_app.valown_address,
                    valad_id=valad_app.app_id,
                    valman=valman,
                    noticeboard_client=valad_app.notbd_client
                )
                return VALAD_NOT_READY_STATUS_CHANGE_OK
            except AttributeError as e:
                logger.log_set_valad_ready_attribute_error(app_id=valad_app.app_id)
                return VALAD_NOT_READY_STATUS_ATTRIBUTE_ERROR
        return VALAD_NOT_READY_STATUS_NO_CHANGE


    def maintain_delcos(
        self
    ):
        """Maintain delegator contracts.
        """
        self.logger.log_maintaining_delcos(num_of_delcos=len(self.delco_app_list.get_app_list()))
        for delco_app in self.delco_app_list.get_app_list():
            self.maintain_single_delco(
                self.algorand_client,
                self.valman,
                copy.copy(delco_app),
                self.partkey_manager,
                self.logger,
            )
            if delco_app.state[0] >> 4 or delco_app.state[0] >> 5:
                self.delco_app_list.remove_single_app(delco_app.app_id)
                self.logger.log_removed_ended_or_deleted_delco(app_id=delco_app.app_id)

        
    @staticmethod
    def maintain_single_delco(
        algorand_client: AlgorandClient,
        valman: AddressAndSigner,
        delco_app: DelcoAppWrapper,
        partkey_manager: PartkeyManager,
        logger: Logger,
    ):
        """Maintain a single delegator contract.

        Parameters
        ----------
        delco_app : DelcoAppWrapper
            Delegator contract app wrapper.
        """
        try:
            delco_state = delco_app.state
            logger.log_state_of_delco_with_id(app_id=delco_app.app_id, state=delco_state)
            if delco_state == DELCO_STATE_READY: # Generate and submit keys (add internal state)
                Daemon.delco_ready_state_handler(
                    algorand_client,
                    valman,
                    delco_app,
                    partkey_manager,
                    logger,
                )
            elif delco_state == DELCO_STATE_SUBMITTED:
                Daemon.delco_submitted_state_handler(
                    algorand_client,
                    valman,
                    delco_app,
                    logger
                )
            elif delco_state == DELCO_STATE_LIVE:
                Daemon.delco_live_state_handler(
                    algorand_client,
                    valman,
                    delco_app,
                    logger
                )
            # Ended contracts no longer visible from validator ad's list -> no need for handler
            elif delco_state[0] >> 4: # Ended contract
                Daemon.delco_ended_state_handler(
                    algorand_client,
                    partkey_manager,
                    delco_app,
                    logger
                )
            elif delco_state[0] >> 5: # Deleted contract
                Daemon.delco_deleted_state_handler(
                    algorand_client,
                    partkey_manager,
                    logger
                )
            else:
                logger.log_unknown_delco_state(state=delco_state)
        except Exception as e:
            logger.error(f'Error while handling delco with ID {delco_app.app_id}: {e}.')


    @staticmethod
    def delco_ready_state_handler(
        algorand_client: AlgorandClient,
        valman: AddressAndSigner,
        delco_app: DelcoAppWrapper,
        partkey_manager: PartkeyManager,
        logger: Logger,
    ):
        """Handle a ready delegator contract (generate and submit keys).

        Parameters
        ----------
        delco_app : DelcoAppWrapper

        Returns
        -------
        int
            Internal state indicator.
        """
        logger.log_delco_in_ready_handler(app_id=delco_app.app_id)
        try: # Check if can pay (funds not frozen)
            report_delco_breach_pay( # Error if can pay, success if can not
                algorand_client=algorand_client,
                valown_address=delco_app.valown_address,
                delman_address=delco_app.delman_address,
                valad_id=delco_app.valad_id,
                delco_id=delco_app.app_id,
                fee_asset_id=delco_app.fee_asa_id,
                valman=valman,
                noticeboard_client=delco_app.notbd_client
            )
            logger.log_delco_cannot_pay(app_id=delco_app.app_id)
            return DELCO_READY_STATUS_BREACH_PAY
        except AttributeError:
            logger.log_attributeerror_cannot_pay(app_id=delco_app.app_id)
        except LogicError: # Failed transaction on LocalNet
            logger.log_logicerror_cannot_pay(app_id=delco_app.app_id)
        except HTTPError: # Failed transaction on public network
            logger.log_httperror_cannot_pay(app_id=delco_app.app_id)
        except AlgodHTTPError: # Failed transaction on public network
            logger.log_httperror_cannot_pay(app_id=delco_app.app_id)
            # Proceed to possible self-reporting
        try: # Check if can pay (funds not frozen)
            report_partkeys_not_submitted( # Error if still time
                algorand_client=algorand_client,
                valown_address=delco_app.valown_address,
                delman_address=delco_app.delman_address,
                valad_id=delco_app.valad_id,
                delco_id=delco_app.app_id,
                fee_asset_id=delco_app.fee_asa_id,
                valman=valman,
                noticeboard_client=delco_app.notbd_client
            )
            logger.log_partkeys_not_submitted(app_id=delco_app.app_id)
            return DELCO_READY_STATUS_NOT_SUBMITTED
        except AttributeError:
            logger.log_attributeerror_partkeys_not_submitted(app_id=delco_app.app_id)
        except LogicError: # Failed transaction on LocalNet
            logger.log_logicerror_partkeys_not_submitted(app_id=delco_app.app_id)
        except HTTPError: # Failed transaction on public network
            logger.log_httperror_partkeys_not_submitted(app_id=delco_app.app_id)
        except AlgodHTTPError: # Failed transaction on public network
            logger.log_httperror_cannot_pay(app_id=delco_app.app_id)
        # Check if partkeys already generated / exist
        try:
            is_generated = partkey_manager.is_partkey_generated(
                address=delco_app.delben_address,
                vote_first_valid=delco_app.round_start,
                vote_last_valid=delco_app.round_end
            )
        except URLError:
            logger.log_urlerror_checking_partkey_generated(app_id=delco_app.app_id)
            return DELCO_READY_STATUS_URL_ERROR
        if is_generated:    # Partkeys generated, submit and move to next state
            logger.log_partkeys_generated_for_delco(app_id=delco_app.app_id)
            # Get key parameters
            partkey_params = partkey_manager.get_existing_partkey_parameters(
                address=delco_app.delben_address,
                vote_first_valid=delco_app.round_start,
                vote_last_valid=delco_app.round_end
            )
            try: # Submit partkeys
                submit_partkeys(
                    algorand_client=algorand_client,
                    valown_address=delco_app.valown_address,
                    delman_address=delco_app.delman_address,
                    delben_address=delco_app.delben_address,
                    partner_address=delco_app.partner_address,
                    valad_id=delco_app.valad_id,
                    delco_id=delco_app.app_id,
                    fee_asset_id=delco_app.fee_asa_id,
                    valman=valman,
                    noticeboard_client=delco_app.notbd_client,
                    vote_first=partkey_params['vote-first-valid'],
                    vote_last=partkey_params['vote-last-valid'],
                    vote_key_dilution=partkey_params['vote-key-dilution'],
                    vote_pk=partkey_params['vote-participation-key'],
                    selection_pk=partkey_params['selection-participation-key'],
                    state_proof_pk=partkey_params['state-proof-key']
                )
                logger.log_partkey_params_submitted(app_id=delco_app.app_id)
                return DELCO_READY_STATUS_SUBMITTED
            except AttributeError:
                logger.log_attributeerror_partkey_submit(app_id=delco_app.app_id)
                return DELCO_READY_STATUS_URL_ERROR
        # Partkey was not generated, so check if it is still pending or if it needs to be generated
        try:
            is_pending = partkey_manager.is_partkey_generation_pending(
                address=delco_app.delben_address,
                vote_first_valid=delco_app.round_start,
                vote_last_valid=delco_app.round_end
            )
        except URLError:
            logger.log_urlerror_checking_partkey_pending(app_id=delco_app.app_id)
            return DELCO_READY_STATUS_URL_ERROR
        if is_pending:      # Generating, wait it out
            logger.log_partkey_generation_pending(app_id=delco_app.app_id)
            return DELCO_READY_STATUS_PENDING
        else:               # Request partkey generation
            res = partkey_manager.add_partkey_generation_request( # No on-chain operation
                address=delco_app.delben_address,
                vote_first_valid=delco_app.round_start,
                vote_last_valid=delco_app.round_end
            )
            if res == PARTKEY_GENERATION_REQUEST_OK_ADDED:
                logger.log_requested_partkey_generation(delco_app.app_id)
                return DELCO_READY_STATUS_REQUESTED
            else:
                logger.log_partkey_generation_denied(delco_app.app_id)
                return DELCO_READY_STATUS_REQUEST_DENIED


    @staticmethod
    def delco_submitted_state_handler(
        algorand_client: AlgorandClient,
        valman: AddressAndSigner,
        delco_app: DelcoAppWrapper,
        logger: Logger,
    ):
        """Handle a delegator contract with submitted keys (report unconfirmed keys).

        Parameters
        ----------
        delco_app : DelcoAppWrapper
        """
        logger.debug(f'In submitted state handler for delco with ID {delco_app.app_id}.')
        try:
            report_unconfirmed_partkeys(
                algorand_client=algorand_client,
                valown_address=delco_app.valown_address,
                delman_address=delco_app.delman_address,
                partner_address=delco_app.partner_address,
                valad_id=delco_app.valad_id,
                delco_id=delco_app.app_id,
                fee_asset_id=delco_app.fee_asa_id,
                valman=valman,
                noticeboard_client=delco_app.notbd_client,
            )
            logger.log_partkeys_not_confirmed(delco_app.app_id)
        except AttributeError:
            logger.log_attributeerror_partkeys_not_confirmed(delco_app.app_id)
        except LogicError:
            logger.log_logicerror_partkeys_not_confirmed(delco_app.app_id)
        except HTTPError: # Failed transaction on public network
            logger.log_httperror_partkeys_not_confirmed(app_id=delco_app.app_id)
        except AlgodHTTPError: # Failed transaction on public network
            logger.log_httperror_cannot_pay(app_id=delco_app.app_id)


    @staticmethod
    def delco_live_state_handler(
        algorand_client: AlgorandClient,
        valman: AddressAndSigner,
        delco_app: DelcoAppWrapper,
        logger: Logger,
    ):
        """Handle a live delegator contract (check different limit breaches and expiry).

        Parameters
        ----------
        delco_app : DelcoAppWrapper
        """
        logger.log_delco_in_live_handler(app_id=delco_app.app_id)
        # Call the coresponding checkup functions, which could result in a state change
        # Check if expired
        try:
            report_contract_expired(
                algorand_client=algorand_client,
                valown_address=delco_app.valown_address,
                delman_address=delco_app.delman_address,
                partner_address=delco_app.partner_address,
                valad_id=delco_app.valad_id,
                delco_id=delco_app.app_id,
                fee_asset_id=delco_app.fee_asa_id,
                valman=valman,
                noticeboard_client=delco_app.notbd_client
            )
            logger.log_contract_expired(delco_app.app_id)
            return
        except AttributeError:
            logger.log_expired_attribute_error(app_id=delco_app.app_id)
        except LogicError:
            logger.log_tried_contract_expired(app_id=delco_app.app_id)
        except HTTPError: # Failed transaction on public network
            logger.log_httperror_contract_expired(app_id=delco_app.app_id)
        except AlgodHTTPError: # Failed transaction on public network
            logger.log_httperror_cannot_pay(app_id=delco_app.app_id)
        # Check if ALGO (max) or gating ASA (min) limits brached
        try:
            report_delben_breach_limits(
                algorand_client=algorand_client,
                valown_address=delco_app.valown_address,
                delman_address=delco_app.delman_address,
                delben_address=delco_app.delben_address,
                partner_address=delco_app.partner_address,
                valad_id=delco_app.valad_id,
                delco_id=delco_app.app_id,
                fee_asset_id=delco_app.fee_asa_id,
                gating_asa_id_list=delco_app.gating_asa_id_list,
                valman=valman,
                noticeboard_client=delco_app.notbd_client
            )
            logger.log_gating_or_stake_limit_breached(app_id=delco_app.app_id)
            return
        except AttributeError:
            logger.log_gating_or_stake_limit_breached_attribute_error(app_id=delco_app.app_id)
        except LogicError:
            logger.log_logicerror_gating_or_stake_limit_breached(app_id=delco_app.app_id)
        except HTTPError: # Failed transaction on public network
            logger.log_httperror_gating_or_stake_limit_breached(app_id=delco_app.app_id)
        except AlgodHTTPError: # Failed transaction on public network
            logger.log_httperror_cannot_pay(app_id=delco_app.app_id)
        try: # Check if can pay
            report_delco_breach_pay( # Error if can pay, success if can not
                algorand_client=algorand_client,
                valown_address=delco_app.valown_address,
                delman_address=delco_app.delman_address,
                valad_id=delco_app.valad_id,
                delco_id=delco_app.app_id,
                fee_asset_id=delco_app.fee_asa_id,
                valman=valman,
                noticeboard_client=delco_app.notbd_client
            )
            logger.log_delco_cannot_pay(app_id=delco_app.app_id)
            return
        except AttributeError:
            logger.log_attributeerror_cannot_pay(app_id=delco_app.app_id)
        except LogicError:
            logger.log_logicerror_cannot_pay(app_id=delco_app.app_id)
        except HTTPError: # Failed transaction on public network
            logger.log_httperror_cannot_pay(app_id=delco_app.app_id)
        except AlgodHTTPError: # Failed transaction on public network
            logger.log_httperror_cannot_pay(app_id=delco_app.app_id)


    @staticmethod
    def delco_ended_state_handler(
        algorand_client: AlgorandClient,
        partkey_manager: PartkeyManager,
        delco_app: DelcoAppWrapper,
        logger: Logger,
    ):
        """Handle an ended delegator contract (delete partkeys).

        Parameters
        ----------
        algorand_client : AlgorandClient
        partkey_manager : PartkeyManager
        delco_app : DelcoAppWrapper
        """
        # logger.debug(f'In ended state handler for delco with ID {delco_app.app_id}.')
        logger.log_delco_in_ended_handler(app_id=delco_app.app_id)
        # Check if the partkey still exists (could be deleted at contract expiry or manually)
        try:
            is_generated = partkey_manager.is_partkey_generated(
                address=delco_app.delben_address,
                vote_first_valid=delco_app.round_start,
                vote_last_valid=delco_app.round_end
            )
        except URLError:
            logger.log_urlerror_checking_partkey_generated(app_id=delco_app.app_id)
        # Scheduled deletion of partkey, which will no longer be used
        if is_generated:
            current_round = algorand_client.client.algod.status()['last-round']
            # Extend duration for the 320 rounds (15 min)
            # To prevent the delayed staking on extension
            # Extension falls under withdrawal
            # Classic withdrawal grants these rounds for free 
            if delco_app.state == DELCO_STATE_ENDED_WITHDREW:
                target_scheduled_deletion = current_round + 320
            else:
                target_scheduled_deletion = current_round
            # Update accordingly
            partkey_manager.update_generated_partkey_scheduled_deletion(
                address=delco_app.delben_address,
                vote_first_valid=delco_app.round_start,
                vote_last_valid=delco_app.round_end,
                scheduled_deletion=target_scheduled_deletion 
            )
            logger.log_scheduled_partkey_deletion_for_ended_or_deleted(
                app_id=delco_app.app_id,
                scheduled_deletion=target_scheduled_deletion,
                round_end=delco_app.round_end
            )
        else:
            logger.log_no_partkeys_found_for_ended_or_deleted(app_id=delco_app.app_id)


    @staticmethod
    def delco_deleted_state_handler(
        algorand_client: AlgorandClient,
        partkey_manager: PartkeyManager,
        delco_app: DelcoAppWrapper,
        logger: Logger,
    ):
        """Handle an ended delegator contract (delete partkeys).

        Parameters
        ----------
        partkey_manager : PartkeyManager
        delco_app : DelcoAppWrapper
        """
        logger.log_delco_in_deleted_handler(app_id=delco_app.app_id)
        # Check if the partkey still exists (could be deleted at contract expiry or manually)
        try:
            is_generated = partkey_manager.is_partkey_generated(
                address=delco_app.delben_address,
                vote_first_valid=delco_app.round_start,
                vote_last_valid=delco_app.round_end
            )
        except URLError:
            logger.debug(f'URL error while checking if partkey already generated for delco with ID {delco_app.app_id}.')
            logger.log_urlerror_checking_partkey_generated(app_id=delco_app.app_id)
        # Scheduled deletion of partkey, which will no longer be used
        if is_generated:
            current_round = algorand_client.client.algod.status()['last-round']
            target_scheduled_deletion = current_round
            partkey_manager.update_generated_partkey_scheduled_deletion(
                address=delco_app.delben_address,
                vote_first_valid=delco_app.round_start,
                vote_last_valid=delco_app.round_end,
                scheduled_deletion=target_scheduled_deletion
            )
            logger.log_scheduled_partkey_deletion_for_ended_or_deleted(
                app_id=delco_app.app_id,
                scheduled_deletion=target_scheduled_deletion,
                round_end=delco_app.round_end
            )
        else:
            logger.log_no_partkeys_found_for_ended_or_deleted(app_id=delco_app.app_id)


    def maintain_contracts(
        self
    ) -> None:
        """Maintain (service) the validator ads and delegator contracts.
        """
        ### Valad part #################################################################################################
        # Read in latest valad app IDs
        self.daemon_config.read_swap()
        # Fetch valads based on the info provided in the config
        valad_id_list = copy.copy(self.daemon_config.validator_ad_id_list)
        self.logger.log_num_of_valad_ids_found(num_of_valads=len(valad_id_list))
        # Add new valads (if any after the first call)
        num_of_added_valads = self.valad_app_list.add_multiple_apps(valad_id_list)
        self.logger.log_num_of_valad_clients_connected(num_of_valads=num_of_added_valads)
        # Update the dynamic information of apps
        num_of_valads, num_of_updated_valads = self.valad_app_list.update_apps()
        self.logger.log_num_of_updated_valads(
            num_of_updated_valads=num_of_updated_valads, 
            num_of_valads=num_of_valads
        )
        # Stop if there will be nothing to service
        if num_of_valads == 0:
            self.logger.log_zero_valad_clients(valad_id_list=valad_id_list)
            return
        # Set to valad ready where applicable
        self.maintain_valads()
        ### Delco part #################################################################################################
        # Fetch IDs of delcos associated with the valads
        delco_id_list = [d_id for valad_app in self.valad_app_list.get_app_list() for d_id in valad_app.delco_id_list]
        self.logger.log_num_of_connected_delcos(num_of_delcos=len(delco_id_list))
        # Add new delcos (if any after the first call)
        num_of_added_delcos = self.delco_app_list.add_multiple_apps(delco_id_list)
        self.logger.log_num_of_delco_clients_connected(num_of_delcos=num_of_added_delcos)
        # Update the dynamic information of apps
        num_of_delcos, num_of_updated_delcos = self.delco_app_list.update_apps()
        self.logger.log_num_of_updated_delcos(
            num_of_updated_delcos=num_of_updated_delcos, 
            num_of_delcos=num_of_delcos
        )
        # Handle each delco based on its state
        self.maintain_delcos()


    def run(
        self
    ) -> None:
        """Run the daemon, periodically calling the smart contract record function.
        """
        while not self.stop_flag:
            # Measure time
            start_time_s = time.time()
            # Check algod
            algod_status = Daemon.check_algod_status(self.algorand_client)
            # If good, run daemon logic
            if algod_status.is_ok:
                self.logger.log_algod_ok_continuing()
                self.logger.log_current_round(current_round=self.algorand_client.client.algod.status()["last-round"])
                try:
                    self.maintain_contracts()        # Service valads and delcos
                except Exception as e:
                    self.logger.log_generic_contract_servicing_error(e)
                try:
                    self.partkey_manager.refresh()  # Manage partkeys
                except Exception as e:
                    self.logger.log_generic_partkey_manager_error(e)
            # If not, report critical
            else:
                self.logger.log_algod_error(algod_status.message)
            # Sleep for the remaining time
            try:
                self.logger.log_single_loop_execution_time(round(time.time() - start_time_s, 2))
                self.logger.log_targeted_sleep_duration(round(start_time_s + self.loop_period_s - time.time(), 2))
                time.sleep(start_time_s + self.loop_period_s - time.time())
            except Exception as e:
                self.logger.log_could_not_sleep(duration_s=self.loop_period_s, e=e)


    def stop(
        self
    ) -> None:
        """Stop the daemon by altering the stop flag.
        """
        self.stop_flag = True
        self.logger.info('Stopping daemon.')



"""
Run the Valar Daemon. For more information see https://pypi.com.

Options:
  - `--config_path`,  Path to the config file. Defaults to `./daemon.config`.
  - `--log_path`,     Path to the log directory (created if does not exist). Defaults to `./valar-daemon-log`
"""
if __name__ == '__main__':

    import argparse
    from pathlib import Path
    from valar_daemon.Daemon import Daemon

    repo_link = 'https://pypi.com'
    parser = argparse.ArgumentParser(description=
        f"Run the Valar Daemon. For more information see {repo_link}."
    )
    parser.add_argument(
        '--config_path', type=str, required=False, 
        help='Path to the config file. Defaults to `./daemon.config`.',
        default=Path(Path.cwd(), 'daemon.config')
    )
    parser.add_argument(
        '--log_path', type=str, required=False, 
        help='Path to the log directory (created if does not exist). Defaults to `./valar-daemon-log`',
        default=Path(Path.cwd(), 'valar-daemon-log')
    )
    args = parser.parse_args()

    print(
        '\n'
        'Pointing the Valar Daemon to the following:\n'
        f'\t Config file at: {args.config_path}\n'
        f'\t Log directory at: {args.log_path}\n'
        '\n'
        'Starting Valar Daemon. Expect no further stdout stream - check the above log directory for feedback.'
        '\n'
    )

    daemon = Daemon(
        args.log_path,
        args.config_path
    )

    daemon.run()



"""Participation key manager definition.

Potential danger points:
- If the output format of partkey gets changed, `get_existing_partkey_parameters` would need adapting.
- General algod API changes (paths, parameters, behavior).
"""
import time
from typing import Tuple

from algokit_utils.beta.algorand_client import AlgorandClient
from algosdk.error import AlgodHTTPError

from valar_daemon.Logger import Logger


"""Return values when requesting partkey generation/
PARTKEY_GENERATION_REQUEST_OK_ADDED              - Request successfully added
PARTKEY_GENERATION_REQUEST_FAIL_IN_PENDING       - Already in pending buffer
PARTKEY_GENERATION_REQUEST_FAIL_IN_GENERATED     - Already generated 
PARTKEY_GENERATION_REQUEST_FAIL_PENDING_FULL     - Pending buffer full
PARTKEY_GENERATION_REQUEST_FAIL_GENERATED_FULL   - Generated buffer full
PARTKEY_GENERATION_REQUEST_FAIL_ALGOD_ERROR      - Encountered Algod error
PARTKEY_GENERATION_REQUEST_FAIL_IN_THE_PAST      - Key validity in the past
"""
PARTKEY_GENERATION_REQUEST_OK_ADDED              =  0
PARTKEY_GENERATION_REQUEST_FAIL_IN_PENDING       = -1
PARTKEY_GENERATION_REQUEST_FAIL_IN_GENERATED     = -2
PARTKEY_GENERATION_REQUEST_FAIL_PENDING_FULL     = -3
PARTKEY_GENERATION_REQUEST_FAIL_GENERATED_FULL   = -4
PARTKEY_GENERATION_REQUEST_FAIL_ALGOD_ERROR      = -5
PARTKEY_GENERATION_REQUEST_FAIL_IN_THE_PAST      = -6


def create_partkey_dict(
    address: str,
    vote_first_valid: int,
    vote_last_valid: int,
    selection_participation_key: str = None,
    state_proof_key: str = None,
    vote_participation_key: str = None,
    vote_key_dilution: str = None,
    id: str=None,
    scheduled_deletion: int=None
) -> dict:
    """Structure partkey information into dictionary.

    Notes
    -----
    Only the address and first/last round is mandatory to facilitate future keys.

    Parameters
    ----------
    address : str
        Account to which partkey belongs.
    vote_first_valid : int
        Starting round.
    vote_last_valid : int
        End round.
    selection_participation_key : str, optional
        Selection key parameter, by default None
    state_proof_key : str, optional
        State proof key parameter, by default None
    vote_participation_key : str, optional
        Participation key parameter, by default None
    vote_key_dilution : str, optional
        Dilution key parameter, by default None
    id : str, optional
        The key's ID, by default None
    scheduled_deletion : int, optional
        Round at which key will be deleted, by default None

    Returns
    -------
    dict
        Structured key information.
    """
    return {
        'address': address,
        'id': id,
        'scheduled-deletion': scheduled_deletion,
        'selection-participation-key': selection_participation_key,
        'state-proof-key': state_proof_key,
        'vote-first-valid': vote_first_valid,
        'vote-key-dilution': vote_key_dilution,
        'vote-last-valid': vote_last_valid,
        'vote-participation-key': vote_participation_key,
    }


class PartkeyBuffer(object):
    """Buffer housing participation key information.
    """

    def __init__(self, max_num_of_keys: int=50):
        """Initialize a partkey buffer.

        Parameters
        ----------
        max_num_of_keys : int, optional
            Maximum number of keys stored in the buffer, by default 50.
        """
        self.max_num_of_keys = max_num_of_keys
        self.partkeys = []

    def is_full(self) -> bool:
        """Check if partkey buffer is full.

        Returns
        -------
        bool
            True if full.
        """
        return len(self.partkeys) >= self.max_num_of_keys
        
    def add_partkey_to_buffer(            
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int,
        selection_participation_key: str=None,
        state_proof_key: str=None,
        vote_participation_key: str=None,
        vote_key_dilution: str=None,
        id=None,
        scheduled_deletion=None
    ):
        """Add a participation key to the buffer.

        Parameters
        ----------
        address : str
            Account to which partkey belongs.
        vote_first_valid : int
            Starting round.
        vote_last_valid : int
            End round.
        selection_participation_key : str, optional
            Selection key parameter, by default None
        state_proof_key : str, optional
            State proof key parameter, by default None
        vote_participation_key : str, optional
            Participation key parameter, by default None
        vote_key_dilution : str, optional
            Dilution key parameter, by default None
        id : str, optional
            The key's ID, by default None
        scheduled_deletion : int, optional
            Round at which key will be deleted, by default None
        """
        if self.is_full():
            raise RuntimeError(f'Partkey buffer is full ({len(self.partkeys)} partkeys present).')
        partkey = create_partkey_dict(
            address=address,
            vote_first_valid=vote_first_valid,
            vote_last_valid=vote_last_valid,
            selection_participation_key=selection_participation_key,
            state_proof_key=state_proof_key,
            vote_participation_key=vote_participation_key,
            vote_key_dilution=vote_key_dilution,
            id=id,
            scheduled_deletion=scheduled_deletion
        )
        self.partkeys.append(partkey)
    
    def pop_partkey_from_buffer(
        self,
        idx: int=0
    ) -> dict:
        """Pop oldest entry.

        Parameters
        ----------
        idx : int, optional
            _description_, by default 0

        Returns
        -------
        dict
            _description_
        """
        return self.partkeys.pop(idx)

    def get_next(
        self
    ) -> dict:
        """Get the next partkey entry.

        Returns
        -------
        dict
            Partkey info.
        """
        if self.is_empty():
            return None
        else:
            return self.partkeys[0]

    def is_empty(
        self
    ) -> bool:
        """Check if buffer is empty.

        Returns
        -------
        bool
            Flag on emptiness of buffer.
        """
        if len(self.partkeys) == 0:
            return True
        else: 
            return False

    def is_partkey_in_buffer(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int
    ) -> bool:
        """Check if the partkey is already in the given buffer.

        Parameters
        ----------
        address : str
            Address of participating entity.
        vote_first_valid : int
            First round when the partkeys are valid.
        vote_last_valid : int
            Last round when the partkeys are valid.
        checked_buffer : List[dict]
            The buffer that is checked.

        Returns
        -------
        bool
        """
        if len(self.partkeys) > 0:
            for entry in self.partkeys:
                if all([
                    entry['address'] == address,
                    entry['vote-first-valid'] == vote_first_valid,
                    entry['vote-last-valid'] == vote_last_valid
                ]):
                    return True
        return False
                    
    def return_partkeys(
            self
    ) -> list:
        """Return a list of the participation keys.

        Returns
        -------
        list
            Participation keys.
        """
        return self.partkeys
    
    def get_index_of_partkey_in_buffer(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int,
    ) -> int:
        """Get the index of a partkey in the buffer.

        Parameters
        ----------
        address : str
            Address of participating entity.
        vote_first_valid : int
            First round when the partkeys are valid.
        vote_last_valid : int
            Last round when the partkeys are valid.

        Returns
        -------
        int
            The index.

        Raises
        ------
        RuntimeError
            Cannot find partkey with ID
        """
        for idx, entry in enumerate(self.partkeys):
            if all([
                entry['address'] == address,
                entry['vote-first-valid'] == vote_first_valid,
                entry['vote-last-valid'] == vote_last_valid
            ]):
                return idx
        raise RuntimeError(f'Cannot find partkey with the requested params for address {address} in buffer')
    
    def update_partkey_scheduled_deletion(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int,
        scheduled_deletion: int
    ) -> Tuple[str, int]:
        """Update the scheduled deletion time.

        Parameters
        ----------
        address : str
            Address of participating entity.
        vote_first_valid : int
            First round when the partkeys are valid.
        vote_last_valid : int
            Last round when the partkeys are valid.
        scheduled_deletion : int
            Round at which the partkey should be deleted

        Returns
        -------
        Tuple(str, int)
            Confirmation of the ID and scheduled deletion round.
        """
        idx = self.get_index_of_partkey_in_buffer(
            address=address,
            vote_first_valid=vote_first_valid,
            vote_last_valid=vote_last_valid,
        )
        self.partkeys[idx]['scheduled-deletion'] = scheduled_deletion
        return (self.partkeys[idx]['id'], self.partkeys[idx]['scheduled-deletion'])


class PartkeyManager(object):
    """Participation key manager.
    """

    busy_msg = 'participation key generation already in progress'


    def __init__(
        self, 
        logger: Logger,
        algorand_client: AlgorandClient
    ):
        self.logger = logger
        self.algorand_client = algorand_client
        self.buffer_pending = PartkeyBuffer()
        self.buffer_generated = PartkeyBuffer()
        self.busy_generating_partkey = False # Flag


    def add_partkey_generation_request(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int,
        vote_key_dilution: int=None,
        scheduled_deletion: int=None
    ):
        """Add partkey generation request to buffer. 

        Parameters
        ----------
        address : str
            Address of participating entity.
        vote_first_valid : int
            First round when the partkeys are valid.
        vote_last_valid : int
            Last round when the partkeys are valid.
        vote_key_dilution : str, optional
            Dilution key parameter, by default None
        scheduled_deletion : int, optional
            Round at which key will be deleted, by default None

        Returns
        -------
        int
            Status code.
        """
        # self.logger.debug(f'Requesting partkey generation between rounds {vote_first_valid} and {vote_last_valid} for address {address}.')
        self.logger.log_partkey_generation_request(
            address=address,
            vote_first_valid=vote_first_valid,
            vote_last_valid=vote_last_valid
        )
        try:
            last_round = self.algorand_client.client.algod.status()['last-round']
        except Exception as e:
            self.logger.log_generic_algod_error(e)
            return PARTKEY_GENERATION_REQUEST_FAIL_ALGOD_ERROR
        # First check if the partkey's validity is in the past
        if last_round >= vote_last_valid:
            self.logger.log_requested_partkey_in_past(num_of_keys=len(self.buffer_pending.partkeys))
            return PARTKEY_GENERATION_REQUEST_FAIL_IN_THE_PAST
        # Check for full buffers
        if self.buffer_pending.is_full():
            # self.logger.debug(f'Pending buffer is full ({self.buffer_pending.max_num_of_keys} partkeys).')
            self.logger.log_pending_buffer_is_full(num_of_keys=len(self.buffer_pending.partkeys))
            return PARTKEY_GENERATION_REQUEST_FAIL_PENDING_FULL
        elif self.buffer_generated.is_full():
            # self.logger.debug(f'Generated buffer is full ({self.buffer_generated.max_num_of_keys} partkeys).')
            self.logger.log_generated_buffer_is_full(num_of_keys=len(self.buffer_generated.partkeys))
            return PARTKEY_GENERATION_REQUEST_FAIL_GENERATED_FULL
        else:
            # Then check if already present in the buffers
            in_pending_buffer = self.is_partkey_generation_pending(
                address, 
                vote_first_valid, 
                vote_last_valid
            )
            in_generated_buffer = self.buffer_generated.is_partkey_in_buffer(
                address, 
                vote_first_valid, 
                vote_last_valid,
            )
            if in_pending_buffer:
                # self.logger.debug(f'Requested partkey already in pending buffer.')
                self.logger.log_requested_partkey_in_pending()
                return PARTKEY_GENERATION_REQUEST_FAIL_IN_PENDING
            elif in_generated_buffer:
                # self.logger.debug(f'Requested partkey already in generated buffer.')
                self.logger.log_requested_partkey_in_generated()
                return PARTKEY_GENERATION_REQUEST_FAIL_IN_GENERATED
            else:
                # Finally, update scheduled deletion and make the request
                if scheduled_deletion is None:
                    scheduled_deletion = vote_last_valid # No early deletion by default
                self.buffer_pending.add_partkey_to_buffer(
                    address,
                    vote_first_valid,
                    vote_last_valid,
                    vote_key_dilution=vote_key_dilution,
                    scheduled_deletion=scheduled_deletion
                )
                # self.logger.debug(f'Added to pending buffer.')
                self.logger.log_partkey_generation_request_added()
                return PARTKEY_GENERATION_REQUEST_OK_ADDED
 
                
    def refresh(
        self
    ) -> None:
        """Update buffers and run any pending partkey generation.
        """
        # if not self.busy_generating_partkey:
        # Conduct maintenance - key deletion (and info fetching) can take place even when busy generating
        self.delete_scheduled_partkeys()                # delete scheduled keys from generated buffer
        self.remove_old_entries_in_buffer_generated()   # delete expired keys from generated buffer
        # Keys need to be generated
        if not self.buffer_pending.is_empty(): # Keygen pending or busy
            next_pending = self.buffer_pending.get_next() # Shared between if clauses - should not execute sequentially!
            if self.busy_generating_partkey: # Check if generating finished
                is_generated = self.is_partkey_generated( # Check if the pending task is already done
                    next_pending['address'],
                    next_pending['vote-first-valid'],
                    next_pending['vote-last-valid']
                )
                if is_generated:
                    self.move_next_partkey_to_generated_buffer() # Move to generated buffer
                    self.busy_generating_partkey = False
                    # Warning!
                    # algod remains busy briefly after the key is already generated
                    # This requires a wait period between observing a key and generating a new one
                    time.sleep(0.1)
            else: # If not busy and pending, generate new one
                self.generate_partkey(
                    next_pending['address'],
                    next_pending['vote-first-valid'],
                    next_pending['vote-last-valid'],
                    next_pending['vote-key-dilution']
                )
                self.busy_generating_partkey = True


    def delete_scheduled_partkeys(self):
        """Delete partkeys that have been scheduled for deletion (e.g. on early contract termination).
        """
        current_round = self.algorand_client.client.algod.status()['last-round']
        for entry in self.buffer_generated.return_partkeys():
            if entry['scheduled-deletion'] is not None:
                if current_round > entry['scheduled-deletion']:
                    is_generated = self.is_partkey_generated( # Check if not already deleted in the meantime
                        entry['address'],
                        entry['vote-first-valid'],
                        entry['vote-last-valid']
                    )
                    if is_generated:
                        self.delete_partkey(
                            entry['address'],
                            entry['vote-first-valid'],
                            entry['vote-last-valid']
                        )


    def is_partkey_generation_pending(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int,
    ) -> bool:
        """Check if partkey generation has been requested.

        Parameters
        ----------
        address : str
            Address of participating entity.
        vote_first_valid : int
            First round when the partkeys are valid.
        vote_last_valid : int
            Last round when the partkeys are valid.

        Returns
        -------
        bool
        """
        return self.buffer_pending.is_partkey_in_buffer(
            address, 
            vote_first_valid, 
            vote_last_valid
        )


    def is_partkey_generated(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int
    ) -> bool:
        """Check if the partkey has already been generated.

        Parameters
        ----------
        address : str
            Address of participating entity.
        vote_first_valid : int
            First round when the partkeys are valid.
        vote_last_valid : int
            Last round when the partkeys are valid.

        Returns
        -------
        bool
        """
        res = self.get_existing_partkey_parameters(address, vote_first_valid, vote_last_valid)
        if res is not None:
            return True
        else:
            return False


    def move_next_partkey_to_generated_buffer(
        self
    ) -> None:
        """Move oldest partkey (FIFO) from pending to generated buffer.
        """
        entry = self.buffer_pending.pop_partkey_from_buffer() # Pop oldest entry
        entry = self.get_existing_partkey_parameters( # Retrieve partkey from algod
            entry['address'],
            entry['vote-first-valid'],
            entry['vote-last-valid']
        )
        self.buffer_generated.add_partkey_to_buffer(
            address=entry['address'],
            vote_first_valid=entry['vote-first-valid'],
            vote_last_valid=entry['vote-last-valid'],
            selection_participation_key=entry['selection-participation-key'],
            state_proof_key=entry['state-proof-key'],
            vote_participation_key=entry['vote-participation-key'],
            vote_key_dilution=entry['vote-key-dilution'],
            id=entry['id']
        )


    # def generate_next_partkey(
    #     self
    # ) -> None:
    #     """Generate the next pending participation key if not busy and the buffer is not empty.
    #     """
    #     # if not self.buffer_pending.is_empty():
    #     entry = self.buffer_pending.get_next()
    #     self.generate_partkey(
    #         entry["address"],
    #         entry["vote-first-valid"],
    #         entry["vote-last-valid"],
    #         entry["vote-key-dilution"]
    #     )
    #     self.busy_generating = True


    def generate_partkey(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int,
        vote_key_dilution: int=None
    ) -> int:
        """Try to generate a participation key.

        Notes
        -----
        If `vote_key_dilution` is kept zero, algod makes the dilution sqrt of the duration.

        Parameters
        ----------
        address : str
            Address of participating entity.
        vote_first_valid : int
            First round when the partkeys are valid.
        vote_last_valid : int
            Last round when the partkeys are valid.
        vote_key_dilution : str, optional
            Dilution key parameter, by default None

        Raises
        ------
        AlgodHTTPError

        Returns
        -------
        int
            Partkey generating flag (0=startd, 1=busy).
        """
        params = dict(
            first=vote_first_valid,
            last=vote_last_valid
        )
        if vote_key_dilution is not None:
            params['dilution'] = vote_key_dilution
        try:
            res = self.algorand_client.client.algod.algod_request(
                'POST', 
                f'/participation/generate/{address}',
                params=params
            )
            # self.logger.info(f'Generating partkey for {address} starting at {vote_first_valid} and ending at {vote_last_valid}.')
            self.logger.log_generating_partkeys(
                address=address,
                vote_first_valid=vote_first_valid,
                vote_last_valid=vote_last_valid
            )
            return 0
        except AlgodHTTPError as e:
            # if e.code != 400:
            if self.busy_msg in e.args[0]:
                return 1
            else:
                raise e
        

    def get_existing_partkey_parameters(
        self, 
        address: str,
        vote_first_valid: int,
        vote_last_valid: int
    ) -> dict:
        """Obtain the relevant parameters of an already generated partkey.

        Parameters
        ----------
        address : str
            Address of participating entity.
        vote_first_valid : int
            First round when the partkeys are valid.
        vote_last_valid : int
            Last round when the partkeys are valid.

        Returns
        -------
        dict
            Partkey parameters. None if did not find the partkey for <address>.

        # Raises
        # ------
        # RuntimeError
        #     Did not find the partkey for <address>.
        """
        res = self.algorand_client.client.algod.algod_request(
            'GET', 
            '/participation'
        )
        if res is not None:
            for entry in res:
                if all([
                    entry['address'] == address,
                    entry['key']['vote-first-valid'] == vote_first_valid,
                    entry['key']['vote-last-valid'] == vote_last_valid
                ]):
                    return create_partkey_dict(
                        address=entry['address'],
                        vote_first_valid=entry['key']['vote-first-valid'],
                        vote_last_valid=entry['key']['vote-last-valid'],
                        selection_participation_key=entry['key']['selection-participation-key'],
                        state_proof_key=entry['key']['state-proof-key'],
                        vote_participation_key=entry['key']['vote-participation-key'],
                        vote_key_dilution=entry['key']['vote-key-dilution'],
                        id=entry['id']
                    )
                    # return {key: entry[key] for key in self.relevant_partkey_parameters}
        # raise RuntimeError(f'Did not find the partkey for {address}.')
        return None


    def remove_old_entries_in_buffer_generated(
        self
    ) -> int:
        """Remove partkeys that are no longer in use.

        Returns
        -------
        int
            Number of removed entries.
        """
        num_of_removed = 0 
        last_round = self.algorand_client.client.algod.status()['last-round'] # Get last round
        for idx, entry in enumerate(self.buffer_generated.return_partkeys()): # Iterate over copy
            if entry['vote-last-valid'] < last_round: # Automatically deleted in partkey list
                # self.buffer_generated.pop(idx) # Delete entry
                self.buffer_generated.pop_partkey_from_buffer(idx)
        # for idx, entry in enumerate(self.buffer_generated.return_partkeys()):
        #     is_generated = self.is_partkey_generated(
        #         entry['address'],
        #         entry['vote-first-valid'],
        #         entry['vote-last-valid']
        #     )
        #      # Marked generated, but not in generated buffer => used up or deleted
        #     if not is_generated:
        #         self.buffer_generated.pop_partkey_from_buffer(idx)
        #         num_of_removed += 1
        return num_of_removed


    def fetch_and_add_existing_partkeys_to_buffer_generated(
        self
    ) -> None:
        """Add partkeys that exist on the node to the generated buffer.
        """
        # Get all partkeys on the node
        res = self.algorand_client.client.algod.algod_request(
            'GET', 
            '/participation'
        )
        # Do nothing if there are not partkeys on the node
        if res is None:
            return
        # Drop the partkeys that are already in the generated buffer
        known_partkey_mask = [False for i in range(len(res))] # Assume all new initially (none known)
        for m, fetched_entry in enumerate(res):
            for n, buffered_entry in enumerate(self.buffer_generated.return_partkeys()):
                if all([
                    fetched_entry['address'] == buffered_entry['address'],
                    fetched_entry['key']['vote-first-valid'] == buffered_entry['vote-first-valid'],
                    fetched_entry['key']['vote-last-valid'] == buffered_entry['vote-last-valid']
                ]):
                    known_partkey_mask[m] = True # Mark the ones that are in the buffer
        if all(known_partkey_mask): # No new keys
            return
        # Maintain the unknonw (new) entries
        res = [res[i] for i in range(len(known_partkey_mask)) if not known_partkey_mask[i]]
        # Buffer the partkeys which are on the node, but not in the generated buffer
        for m, entry in enumerate(res):
            self.buffer_generated.add_partkey_to_buffer(
                address=entry['address'],
                vote_first_valid=entry['key']['vote-first-valid'],
                vote_last_valid=entry['key']['vote-last-valid'],
                selection_participation_key=entry['key']['selection-participation-key'],
                state_proof_key=entry['key']['state-proof-key'],
                vote_participation_key=entry['key']['vote-participation-key'],
                vote_key_dilution=entry['key']['vote-key-dilution'],
                id=entry['id']
            )


    def delete_partkey(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int
    ) -> None:
        """Delete the participation key.

        Parameters
        ----------
        address : str
            Address of participating entity.
        vote_first_valid : int
            First round when the partkeys are valid.
        vote_last_valid : int
            Last round when the partkeys are valid.

        Raises
        ------
        RuntimeError
            Tried deleting partkey that is not in generated buffer.
        RuntimeError
            Tried deleting non-existent partkey.
        """
        # Check if in buffer
        in_gen_buf = self.buffer_generated.is_partkey_in_buffer(
            address,
            vote_first_valid,
            vote_last_valid
        )
        if not in_gen_buf:
            raise RuntimeError(f'Tried deleting partkey that is not in generated buffer for address {address}')
        # Check if exists at all
        generated = self.is_partkey_generated(
            address,
            vote_first_valid,
            vote_last_valid
        )
        if not generated:
            raise RuntimeError(f'Tried deleting non-existent partkey for address {address}')
        partkey_id = self.get_partkey_id(
            address,
            vote_first_valid,
            vote_last_valid
        )
        self.delete_partkey_using_id(partkey_id)


    def delete_partkey_using_id(
        self,
        partkey_id: int
    ) -> None:
        """Delete the participation key.

        Parameters
        ----------
        partkey_id : int
            ID of the participation key.
        """
        self.algorand_client.client.algod.algod_request(
            'DELETE', 
            f'/participation/{partkey_id}'
        )


    def get_partkey_id(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int
    ) -> int:
        """Get partkey ID.

        Parameters
        ----------
        address : str
            Address of participating entity.
        vote_first_valid : int
            First round when the partkeys are valid.
        vote_last_valid : int
            Last round when the partkeys are valid.

        Returns
        -------
        int
            Partkey ID.
        """
        return self.get_existing_partkey_parameters(
            address,
            vote_first_valid,
            vote_last_valid
        )['id']


    def update_generated_partkey_scheduled_deletion(
        self,
        address: str,
        vote_first_valid: int,
        vote_last_valid: int,
        scheduled_deletion: int
    ) -> Tuple[str, int]:
        """Update the scheduled deletion time for a generated partkey.

        Parameters
        ----------
        address : str
            Address of participating entity.
        vote_first_valid : int
            First round when the partkeys are valid.
        vote_last_valid : int
            Last round when the partkeys are valid.
        scheduled_deletion : int
            Round at which the partkey should be deleted

        Returns
        -------
        Tuple(str, int)
            Confirmation of the ID and scheduled deletion round.
        """
        return self.buffer_generated.update_partkey_scheduled_deletion(
            address=address,
            vote_first_valid=vote_first_valid,
            vote_last_valid=vote_last_valid,
            scheduled_deletion=scheduled_deletion
        )
